#
# Collective Knowledge (collaborative program optimization using mobile devices (such as Android mobile phones and tables))
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
line='********************************************************************'

fpack='crowd-pack.zip'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# prepare experiments for crowdsourcing using mobile phones

def request(i):
    """
    Input:  {
              (crowd_uid)         - if !='', processing results and possibly chaining experiments

              (email)             - email or person UOA
              (platform_features) - remote device platform features

              (scenario)          - pre-set scenario
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    from random import randint
    import copy
    import shutil
    import zipfile
    import json

    # Setting output
    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    rr={'return':0}

    email=i.get('email','')

    ruoa=i.get('record_repo_uoa','')
#    if ruoa=='': ruoa='upload'
    # Hack
    ck.cfg["forbid_writing_to_local_repo"]="no"
    ck.cfg["allow_writing_only_to_allowed"]="no"
    ck.cfg["forbid_global_delete"]="no"
#    ck.cfg["allow_run_only_from_allowed_repos"]="yes"

    # Check if processing started experiments
    cuid=i.get('crowd_uid','')
    if cuid!='':
       ###################################################################################################################
       # Load info
       r=ck.access({'action':'load',
                    'module_uoa':work['self_module_uid'],
                    'data_uoa':cuid})
       if r['return']>0: return r
       d=r['dict']

       euoa=d['experiment_uoa']
       ol=d['off_line']

       suid=ol.get('solution_uid','') # should normally be prepared in advance!

       scenario_uoa=ol['scenario_module_uoa']
       condition_objective='#'+ol['meta']['objective']

       xstatus=''

       results=i.get('results',{})

       #Log
       r=ck.access({'action':'log',
                    'module_uoa':cfg['module_deps']['experiment'],
                    'text':'Finishing crowd experiment: '+cuid+' ('+email+')\n'+json.dumps(results,indent=2,sort_keys=True)+'\n'})
       if r['return']>0: return r

       if len(results)>0:

          repeat=results.get('ct_repeat','')
          if repeat=='': repeat=1

          cpu_freq0=results.get('cpu_freq0',{})
          cpu_freq1=results.get('cpu_freq1',{})

          ch0=results.get('characteristics0',{})
          ch1=results.get('characteristics1',{})

          # TBD: improve stat analysis -> use CK module (here quick prototyping)
          fch0min=-1
          fch0max=-1
          for q in ch0:
              v=ch0[q]
              if fch0min==-1 or v<fch0min:
                 fch0min=v
              if fch0max==-1 or v>fch0max:
                 fch0max=v

          var=(fch0max-fch0min)/fch0min

          fch1min=-1
          fch1max=-1
          for q in ch1:
              v=ch1[q]
              if fch1min==-1 or v<fch1min:
                 fch1min=v
              if fch1max==-1 or v>fch1max:
                 fch1max=v

          impr=0.0
          if fch1min!=0: impr=fch0min/fch1min

          ol["meta_extra"]["cpu_cur_freq"]=cpu_freq1

          sol=ol["solutions"][0]
          sol["extra_meta"]["cpu_cur_freq"]=cpu_freq1

          point=sol["points"][0]

          point["characteristics"]["##characteristics#run#execution_time_kernel_0#min"]=fch0min
          point["characteristics"]["##characteristics#run#repeat#min"]=repeat

          point["improvements"]["##characteristics#run#execution_time_kernel_0#min_imp"]=impr

#          Hack: don't write for now, otherwise most of the time ignored ...
          var=-1
          point["misc"]["##characteristics#run#execution_time_kernel_0#range_percent"]=var

          sol["points"][0]=point
          ol["solutions"][0]=sol

          # Get conditions from a scenario
          r=ck.access({'action':'load',
                       'module_uoa':cfg['module_deps']['module'],
                       'data_uoa':scenario_uoa})
          if r['return']>0: return r
          ds=r['dict']

          scon=ds.get('solution_conditions',[])
          if len(scon)>0:
             con=copy.deepcopy(point["characteristics"])
             con.update(point["improvements"])
             con.update(point["misc"])
             # Hack
             con["##characteristics#compile#md5_sum#min_imp"]=0

             ii={'action':'check',
                 'module_uoa':cfg['module_deps']['math.conditions'],
                 'new_points':['0'],
                 'results':[{'point_uid':'0', 'flat':con}],
                 'conditions':scon,
                 'middle_key':condition_objective,
                 'out':oo}
             ry=ck.access(ii)
             if ry['return']>0: return ry 

             xdpoints=ry['points_to_delete']
             if len(xdpoints)>0:
                xstatus='*** Your explored solution is not better than existing ones (conditions are not met) ***\n' 
                if o=='con':
                   ck.out('')
                   ck.out('    WARNING: conditions on characteristics were not met!')
             else:
                # Submitting solution
                ii=copy.deepcopy(ol)
                ii['action']='add_solution'
                ii['module_uoa']=cfg['module_deps']['program.optimization']
                ii['repo_uoa']='upload' # Hack 
                ii['user']=email
                rx=ck.access(ii)
                if rx['return']>0: return rx

                if rx.get('recorded','')=='yes':
                   ri=rx.get('recorded_info',{})
                   xstatus=ri.get('status','')
                   xlog=ri.get('log','')

                   rz=ck.access({'action':'log',
                                 'module_uoa':cfg['module_deps']['experiment'],
                                 'file_name':cfg['log_file_results'],
                                 'text':xlog})
                   if rz['return']>0: return rz

                else:
                   xstatus='*** Your explored solution is not better than existing ones ***\n' 

             r=ck.access({'action':'log',
                          'module_uoa':cfg['module_deps']['experiment'],
                          'text':'Result of crowd experiment (UID='+suid+') : '+cuid+' ('+email+'): '+xstatus+'\n'})
             if r['return']>0: return r


       # Cleaning experiment entry
       r=ck.access({'action':'delete',
                    'module_uoa':cfg['module_deps']['experiment'],
                    'data_uoa':euoa})
       if r['return']>0: return r

       # Cleaning crowd entry
       r=ck.access({'action':'delete',
                    'module_uoa':work['self_module_uid'],
                    'data_uoa':cuid})
       if r['return']>0: return r

       # Finishing
       status='Crowdsourced results from your mobile device were successfully processed by Collective Knowledge Aggregator!\n\n'+xstatus

       if o=='con':
          ck.out('')
          ck.out(status)

       rr['status']=status

    else:
       ###################################################################################################################
       # Initialize platform
       pf=i.get('platform_features',{})

       cpu_abi=pf.get('cpu',{}).get('cpu_abi','')
       os_bits=pf.get('os',{}).get('bits','')

       tos=''
       static=''
       max_size_pack=1200000
       extra_tags=''
       if cpu_abi.startswith('armeabi-'):
          tos='android19-arm'
          extra_tags='arm-specific'
       elif cpu_abi.startswith('arm64'):
          tos='android21-arm64'
#          extra_tags='arm-specific'
       elif cpu_abi=='x86':
          tos='android19-x86'
          static='yes'
          max_size_pack=2200000
#          if os_bits=='64':
#             tos='android21-x86_64'

       if tos=='':
          return {'return':1, 'error':'ABI of your mobile device is not yet supported for crowdtuning ('+cpu_abi+') - please contact author (Grigori.Fursin@cTuning.org) to check if it\'s in development'}

       tdid=''
       hos=''

       xscenario=i.get('scenario','')

       scenarios=cfg['scenarios']
       ls=len(scenarios)

       # Prepare platform info
       ii={'action':'detect',
           'module_uoa':cfg['module_deps']['platform.os'],
           'target_os':tos,
           'skip_info_collection':'yes',
           'out':oo}
       pi=ck.access(ii)
       if pi['return']>0: return pi
       del(pi['return'])

       # Merge with remote device platform features
       r=ck.merge_dicts({'dict1':pi['features'], 'dict2':pf})
       if r['return']>0: return r

       pf['features']=r['dict1']

       #Log
       r=ck.dumps_json({'dict':pf, 'skip_indent':'yes', 'sort_keys':'yes'})
       if r['return']>0: return r
       x=r['string']

       r=ck.access({'action':'log',
                    'module_uoa':cfg['module_deps']['experiment'],
                    'text':email+'\n'+x+'\n'})
       if r['return']>0: return r

       # Try to generate at least one experimental pack!
       n=0
       nm=32

       success=False
       while n<nm and not success:
          n+=1

          # select scenario randomly
          if xscenario!='': scenario=xscenario
          else:             scenario=scenarios[randint(0,ls-1)]

          pic=copy.deepcopy(pi)

          ii={'action':'crowdsource',
              'module_uoa':scenario,
              'target_os':tos,
              'local':'yes',
              'quiet':'yes',
              'iterations':1,
              'platform_info':pic,
              'once':'yes',
              'skip_collaborative':'yes',
              'parametric_flags':'yes',
#              'static':'yes',
#              'program_uoa':'*susan',
              'any_flag_tags':extra_tags,
#              'cmd_key':'edges',
#              'dataset_uoa':'image-pgm-0001',
              'extra_dataset_tags':['small'],
              'no_run':'yes',
              'keep_experiments':'yes',
              'new':'yes',
              'static':static,
              'skip_pruning':'yes',
              'skip_info_collection':'yes',
              'out':oo}
          rrr=ck.access(ii)
          if rrr['return']>0:
             if o=='con':
                ck.out('')
                ck.out('WARNING: '+rrr['error']+' - can\'t continue this sub-scenario ...')
                ck.out('')
          else:
             # Prepare pack ...
             ri=rrr.get('recorded_info',{})
             ruid=ri.get('recorded_uid','')

             lio=rrr.get('last_iteration_output',{})
             fail=lio.get('fail','')
             if fail=='yes' or 'off_line' not in rrr: # sometimes off_line not in rrr, why I don't know yet
                if o=='con':
                   ck.out('')
                   ck.out('WARNING: Pipeline failed ('+lio.get('fail_reason','')+')')
                   ck.out('')

                # Delete failed experiment
                if ruid!='':
                   ii={'action':'delete',
                       'module_uoa':cfg['module_deps']['experiment'],
                       'data_uoa':ruid}
                   r=ck.access(ii)
                   # ignore output

             else:
                # Prepare pack
                ol=rrr['off_line']
                ed=rrr.get('experiment_desc',{})
                choices=ed.get('choices',{})

                d={'experiment_uoa':ruid,
                   'off_line':ol}

                ii={'action':'add',
                    'module_uoa':work['self_module_uid'],
                    'repo_uoa':ruoa,
                    'dict':d}
                r=ck.access(ii)
                if r['return']>0: return r

                p=r['path']

                cuid=r['data_uid'] # crowd experiment identifier
                rr['crowd_uid']=cuid

                x=lio.get('characteristics',{}).get('compile',{}).get('joined_compiler_flags','')

                dsc='Scenario: '+rrr.get('scenario_desc','')+'\n'
                dsc+='Sub-scenario: '+rrr.get('subscenario_desc','')+'\n'
                dsc+='Benchmark/codelet: '+choices.get('data_uoa','')+'\n'
                dsc+='CMD key: '+choices.get('cmd_key','')+'\n'
                dsc+='Dataset: '+choices.get('dataset_uoa','')+'\n'
                dsc+='Dataset file: '+choices.get('dataset_file','')+'\n'
                dsc+='Optimizations:\n'
                dsc+='* OpenCl tuning: not used\n'
                dsc+='* Compiler description: '+choices.get('compiler_description_uoa','')+'\n'
                dsc+='* Compiler flags: -O3 vs '+x+'\n'

                rr['desc']=dsc

                deps=lio.get('dependencies',{})
                for kdp in deps:
                    dp=deps[kdp]
                    z=dp.get('cus',{})
                    dl=z.get('dynamic_lib','')
                    pl=z.get('path_lib','')

                    if dl!='' and pl!='':
                       pidl=os.path.join(pl, dl)
                       if os.path.isfile(pidl):
                          pidl1=os.path.join(p, dl)
                          try:
                             shutil.copyfile(pidl, pidl1)
                          except Exception as e: 
                             pass

                if o=='con':
                   ck.out('')
                   ck.out('  Crowd UID: '+cuid)

                # Copying binaries and inputs here
                target_exe_0=rrr.get('original_target_exe','')
                target_path_0=rrr.get('original_path_exe','')
                target_exe_1=lio.get('state',{}).get('target_exe','')
                tp1=rrr.get('new_path_exe','')

                tp0=os.path.dirname(target_path_0)
                target_path_1=os.path.join(tp0,tp1)

                if o=='con':
                   ck.out('')
                   ck.out('Copying executables:')
                   ck.out(' * '+target_path_0+'  /  '+target_exe_0)
                   ck.out(' * '+target_path_1+'  /  '+target_exe_1)
                   ck.out('')

                duoa=choices.get('dataset_uoa','')
                dfile=choices.get('dataset_file','')

                # create cmd
                prog_uoa=choices.get('data_uoa','')
                cmd_key=choices.get('cmd_key','')
                r=ck.access({'action':'load',
                             'module_uoa':cfg['module_deps']['program'],
                             'data_uoa':prog_uoa})
                if r['return']>0: return r
                dd=r['dict']
                pp=r['path']

                rcm=dd.get('run_cmds','').get(cmd_key,{}).get('run_time',{}).get('run_cmd_main','')
                rcm=rcm.replace('$#BIN_FILE#$ ','')
                rcm=rcm.replace('$#dataset_path#$','')
                rcm=rcm.replace('$#dataset_filename#$',dfile)
                rcm=rcm.replace('$#src_path#$','')

                rif=dd.get('run_cmds','').get(cmd_key,{}).get('run_time',{}).get('run_input_files',[])

                if o=='con':
                   ck.out('Cmd: '+rcm)

                if target_path_0!='' and target_path_1!='' and target_exe_0!='' and target_exe_1!='' and \
                   not (rcm.find('$#')>=0 or rcm.find('#$')>=0 or rcm.find('<')>=0):

                   te0=os.path.join(target_path_0, target_exe_0)
                   te1=os.path.join(target_path_1, target_exe_1)

                   nte0=os.path.join(p, target_exe_0)
                   nte1=os.path.join(p, target_exe_1)

                   # Copying binary files
                   copied=True
                   try:
                      shutil.copyfile(te0, nte0)
                      shutil.copyfile(te1, nte1)

                      for inp in rif:
                          px1=os.path.join(pp, inp)
                          px2=os.path.join(p, inp)
                          shutil.copyfile(px1, px2)

                   except Exception as e: 
                      copied=False
                      pass

                   if copied:
                      # clean dirs
                      try:
                         shutil.rmtree(target_path_0, ignore_errors=True)
                         shutil.rmtree(target_path_1, ignore_errors=True)
                      except Exception as e: 
                         if o=='con':
                            ck.out('')
                            ck.out('WARNING: can\'t fully erase tmp dir')
                            ck.out('')
                         pass

                      if o=='con':
                         ck.out('Copying datasets ...')

                      # Check dataset files
                      rr['choices']=choices

                      copied=True
                      if duoa!='' and dfile!='':
                         r=ck.access({'action':'load',
                                      'module_uoa':cfg['module_deps']['dataset'],
                                      'data_uoa':duoa})
                         if r['return']>0: return r

                         pd=r['path']

                         td=os.path.join(pd, dfile)
                         ntd=os.path.join(p, dfile)

                         copied=True
                         try:
                            shutil.copyfile(td, ntd)
                         except Exception as e: 
                            copied=False
                            pass

                      if copied:
                         if o=='con':
                            ck.out('Preparing zip ...')

                         # Prepare archive
                         zip_method=zipfile.ZIP_DEFLATED

                         gaf=i.get('all','')

                         fl={}

                         r=ck.list_all_files({'path':p})
                         if r['return']>0: return r

                         flx=r['list']

                         for k in flx:
                             fl[k]=flx[k]

                         pfn=os.path.join(p, fpack)

                         # Write archive
                         copied=True
                         try:
                            f=open(pfn, 'wb')
                            z=zipfile.ZipFile(f, 'w', zip_method)
                            for fn in fl:
                                p1=os.path.join(p, fn)
                                z.write(p1, fn, zip_method)
                            z.close()
                            f.close()

                         except Exception as e:
                            copied=False

                         if copied:
                            if o=='con':
                               ck.out('Preparing cmd ...')

                            size=os.path.getsize(pfn) 

                            r=ck.convert_file_to_upload_string({'filename':pfn})
                            if r['return']>0: return r

                            fx=r['file_content_base64']

                            #MD5
                            import hashlib
                            md5=hashlib.md5(fx.encode()).hexdigest()

                            if o=='con':
                               ck.out('Finalizing ...')

                            calibrate='no'
                            if dd.get('run_vars',{}).get('CT_REPEAT_MAIN','')!='':
                               calibrate='yes'

                            if len(fx)>max_size_pack:
                               if o=='con':
                                  ck.out('')
                                  ck.out('WARNING: pack is too large ('+str(len(fx))+')')
                                  ck.out('')
                            else:
                               # finalize info
                               rr['file_content_base64']=fx
                               rr['size']=size 
                               rr['md5sum']=md5
                               rr['run_cmd_main']=rcm
                               rr['bin_file0']=target_exe_0
                               rr['bin_file1']=target_exe_1
                               rr['calibrate']=calibrate
                               rr['calibrate_max_iters']=10
                               rr['calibrate_time']=10.0
                               rr['repeat']=5
                               rr['ct_repeat']=1

                               success=True

                if not success:
                   if o=='con':
                      ck.out('')
                      ck.out('WARNING: some files are missing - removing crowd entry ('+cuid+') ...')

                   ii={'action':'rm',
                       'module_uoa':work['self_module_uid'],
                       'data_uoa':cuid}
                   r=ck.access(ii)
                   if r['return']>0: return r

       if not success:
          rr={'return':1, 'error':'could not create any valid expeirmental pack for your mobile - possibly internal error! Please, contact authors'}

    return rr

##############################################################################
# start server to process tasks

def server(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import time
    import os
    import datetime

    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    try:
        cur_dir=os.getcwd()
    except OSError:
        os.chdir('..')
        cur_dir=os.getcwd()

    # Get path
    ii={'action':'find',
        'module_uoa':cfg['module_deps']['tmp'],
        'data_uoa':cfg['tmp-server']}
    r=ck.access(ii)
    if r['return']>0:
       if r['return']!=16: return r
       ii['action']='add'
       r=ck.access(ii)
       if r['return']>0: return r

    pp=r['path'] 

    fmr=cfg['file-mobile-request']

    while True:
       if o=='con':
          ck.out('Quering for tasks ...')

       os.chdir(cur_dir) # restore path that may be changed during crowd-pack generation

       dirList=os.listdir(pp)
       for q in dirList:
           p=os.path.join(pp,q)
           if os.path.isfile(p) and q.startswith(fmr):
              r=ck.load_json_file({'json_file':p})
              if r['return']==0:
                 ic=r['dict']

                 pr=os.path.join(pp,'result-'+q) # result file

                 xt1=ic.get('iso_datetime_of_request','')

                 r=ck.get_current_date_time({})
                 xt2=r['iso_datetime']

                 # Check that not outdated
                 r=ck.convert_iso_time({'iso_datetime':xt1})
                 if r['return']>0: return r
                 t1=r['datetime_obj']

                 r=ck.convert_iso_time({'iso_datetime':xt2})
                 if r['return']>0: return r
                 t2=r['datetime_obj']

                 dt=(t2-t1).total_seconds()

                 if dt>600:
                    if o=='con':
                       ck.out('Found outdated request (created '+str(dt)+' secs. ago) - removing ...')

                    if os.path.isfile(p):
                       os.remove(p)

                    pf, pff = os.path.split(p)
                    px=os.path.join(pf,'result-',pff)

                    if os.path.isfile(px):
                       os.remove(p)
                 else:

                    if ic.get('status_ongoing','')=='yes' or ic.get('status_finished','')=='yes':
                       continue

                    if o=='con':
                       ck.out('**************************************')
                       ck.out('Found request: '+q)

                    # Updating file
                    ic['status_ongoing']='yes'

                    r=ck.save_json_to_file({'json_file':p, 'dict':ic})
                    if r['return']>0: return r

                    if o=='con':
                       ck.out('Preparing experiment pack ...')

                    ic['out']=oo

                    r=request(ic)

                    if o=='con':
                       ck.out('Updating request file - saying that ready for download !')

                    # Record outcome for the request
                    r=ck.save_json_to_file({'json_file':pr, 'dict':r})
                    if r['return']>0: return r

                    ic['status_finished']='yes'
                    r=ck.save_json_to_file({'json_file':p, 'dict':ic})
                    if r['return']>0: return r

       time.sleep(10)

    return {'return':0}

##############################################################################
# request experiment pack for mobile device

def crowdsource(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import time
    import os
    import copy

    # Test that new version
    if i.get('new_engine','')!='yes':
       return {'return':1, 'error':'your CK application to crowdsource experiments is OUTDATED - update it please!'}

    ic=copy.deepcopy(i)

    o=i.get('out','')

    # Get path
    ii={'action':'find',
        'module_uoa':cfg['module_deps']['tmp'],
        'data_uoa':cfg['tmp-server']}
    r=ck.access(ii)
    if r['return']>0: return r

    pp=r['path'] 

    rx=ck.gen_uid({})
    if rx['return']>0: return rx
    quid=rx['data_uid']

    fmr=cfg['file-mobile-request']+'-'+quid+'.json'
    pf=os.path.join(pp,fmr)

    r=ck.get_current_date_time({})
    ic['iso_datetime_of_request']=r['iso_datetime']

    ll=['cid','out','web','out_file','cids', 'xcids', 'con_encoding', 'action']
    for q in ll:
        if q in ic: del(ic[q])

    r=ck.save_json_to_file({'json_file':pf, 'dict':ic})
    if r['return']>0: return r

    return {'return':0, 'queue_uid':quid}

##############################################################################
# check if crowd-pack is ready

def check(i):
    """
    Input:  {
              queue_uid - check this task (if crowd pack is ready)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    quid=i.get('queue_uid','')
    if quid=='':
       return {'return':1, 'error':'queue_uid is not specified - possibly corrupted or outdated app to crowdsource experiments'}

    cuid=''

    o=i.get('out','')

    # Get path
    ii={'action':'find',
        'module_uoa':cfg['module_deps']['tmp'],
        'data_uoa':cfg['tmp-server']}
    r=ck.access(ii)
    if r['return']>0: return r
    p=r['path']

    rr={'return':0, 'crowd_uid':''}

    ff=cfg['file-mobile-request']+'-'+quid+'.json'
    pp=os.path.join(p,'result-'+ff)
    if os.path.isfile(pp):
       r=ck.load_json_file({'json_file':pp})
       if r['return']==0:
          rr=r['dict']

          # Removing files
          pp1=os.path.join(p,ff)
          if os.path.isfile(pp1):
             os.remove(pp1)

          if os.path.isfile(pp):
             os.remove(pp)

    return rr



#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

import os
cfg = {}  # Will be updated by CK (meta description of this module)
work = {}  # Will be updated by CK (temporal data)
ck = None  # Will be updated by CK (initialized CK kernel)

# Local settings

##############################################################################
# Initialize module


def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return': 0}

##############################################################################
# Add module


def add(i):
    """
    Input:  {
              (repo_uoa)          - repo UOA
              module_uoa          - normally should be 'module' already
              data_uoa            - UOA of the module to be created

              (desc)              - module description
              (license)           - module license
              (copyright)         - module copyright
              (developer)         - module developer
              (developer_email)   - module developer
              (developer_webpage) - module developer
              (actions)           - dict with actions {"func1":{}, "func2":{} ...}
              (actions_redirect)  - dict with actions redrect {"func":"real func", ...}
              (dict)              - other meta description to add to entry

              (quiet)             - minimal interaction
              (func)              - just add one dummy action
              (support_web)       - if 'yes', make it a web API, i.e. allow an access to this function in the CK server
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              Output of the 'add' kernel function
            }

    """

    # Check if global writing is allowed
    r = ck.check_writing({'module_uoa': work['self_module_uoa']})
    if r['return'] > 0:
        return r

    o = i.get('out', '')

    # Find path to module 'module' to get dummies
    r = ck.access({'action': 'load',
                   'module_uoa': work['self_module_uoa'],
                   'data_uoa': work['self_module_uoa'],
                   'common_func': 'yes'})
    if r['return'] > 0:
        return r
    p = r['path']

    pm = os.path.join(p, cfg['dummy_module'])
    pma = os.path.join(p, cfg['dummy_module_action'])

    # Load module dummy
    r = ck.load_text_file({'text_file': pm})
    if r['return'] > 0:
        return r
    spm = r['string']

    # Load module action dummy
    r = ck.load_text_file({'text_file': pma})
    if r['return'] > 0:
        return r
    spma = r['string']

    # Prepare meta description
    desc = i.get('desc', '')
    license = i.get('license', '')
    copyright = i.get('copyright', '')
    developer = i.get('developer', '')
    developer_email = i.get('developer_email', '')
    developer_webpage = i.get('developer_webpage', '')
    actions = i.get('actions', {})
    actions_redirect = i.get('actions_redirect', {})

    func = i.get('func', '')
    if func != '':
        if ck.cfg.get('allowed_action_names', '') != '':
            import re

            anames = ck.cfg.get('allowed_action_names', '')

            if not re.match(anames, func):
                return {'return': 1, 'error': 'found disallowed characters in the action name (allowed: "'+anames+'")'}

        if '-' in func:
            func1 = func.replace('-', '_')
            actions_redirect[func] = func1

        actions[func] = {}

    quiet = i.get('quiet', '')

    # If console mode, ask some questions
    if quiet != 'yes' and o == 'con':
        if desc == '':
            ck.out('')
            r = ck.inp({'text': 'Add brief module description: '})
            desc = r['string']

        if license == '' and ck.cfg.get('default_license', '') != '':
            ck.out('')
            r = ck.inp(
                {'text': 'Add brief module license (or Enter to use "'+ck.cfg['default_license']+'"): '})
            license = r['string']
            if license == '':
                license = ck.cfg['default_license']

        if copyright == '' and ck.cfg.get('default_copyright', '') != '':
            ck.out('')
            r = ck.inp(
                {'text': 'Add brief module copyright (or Enter to use "'+ck.cfg['default_copyright']+'"): '})
            copyright = r['string']
            if copyright == '':
                copyright = ck.cfg['default_copyright']

        if developer == '' and ck.cfg.get('default_developer', '') != '':
            ck.out('')
            r = ck.inp(
                {'text': 'Add module\'s developer (or Enter to use "'+ck.cfg['default_developer']+'"): '})
            developer = r['string']
            if developer == '':
                developer = ck.cfg['default_developer']

        if developer_email == '' and ck.cfg.get('default_developer_email', '') != '':
            ck.out('')
            r = ck.inp({'text': 'Add module\'s developer email (or Enter to use "' +
                        ck.cfg['default_developer_email']+'"): '})
            developer_email = r['string']
            if developer_email == '':
                developer_email = ck.cfg['default_developer_email']

        if developer_webpage == '' and ck.cfg.get('default_developer_webpage', '') != '':
            ck.out('')
            r = ck.inp({'text': 'Add module\'s developer webpage (or Enter to use "' +
                        ck.cfg['default_developer_webpage']+'"): '})
            developer_webpage = r['string']
            if developer_webpage == '':
                developer_webpage = ck.cfg['default_developer_webpage']

        if len(actions) == 0:
            act = '*'
            while act != '':
                ck.out('')

                r = ck.inp(
                    {'text': 'Add action function (or Enter to stop): '})
                act = r['string'].strip()
                if act != '':

                    if ck.cfg.get('allowed_action_names', '') != '':
                        import re

                        anames = ck.cfg.get('allowed_action_names', '')

                        if not re.match(anames, act):
                            return {'return': 1, 'error': 'found disallowed characters in the action name (allowed: "'+anames+'")'}

                    actions[act] = {}

                    if '-' in act:
                        act1 = act.replace('-', '_')
                        actions_redirect[act] = act1

                    fweb = i.get('support_web', '')
                    if fweb != '':
                        if fweb != 'yes':
                            fweb = ''

#                r1=ck.inp({'text':'Support web (y/N): '})
#                x=r1['string'].lower()
#                if x=='yes' or x=='y':
#                   fweb='yes'
                    if fweb != '':
                        actions[act]['for_web'] = fweb

                    r1 = ck.inp({'text': 'Add action description: '})
                    adesc = r1['string']
                    if adesc != '':
                        actions[act]['desc'] = adesc

    ck.out('')

    # Prepare meta description
    dd = {}
    if desc != '':
        dd['desc'] = desc
    spm = spm.replace('$#desc#$', desc)

    if license != '':
        dd['license'] = license
    spm = spm.replace('$#license#$', license)

    if copyright != '':
        dd['copyright'] = copyright
    spm = spm.replace('$#copyright#$', copyright)

    dev = ''
    if developer != '':
        dev = developer
        dd['developer'] = developer

    if developer_email != '':
        if dev != '':
            dev += ', '
        dev += developer_email
        dd['developer_email'] = developer_email

    if developer_webpage != '':
        if dev != '':
            dev += ', '
        dev += developer_webpage
        dd['developer_webpage'] = developer_webpage

    spm = spm.replace('$#developer#$', dev)

    dd['actions'] = actions
    if len(actions_redirect) > 0:
        dd['actions_redirect'] = actions_redirect

    # Substitute actions
    for act in actions:
        adesc = actions[act].get('desc', 'TBD: action description')
        if act in actions_redirect:
            act = actions_redirect[act]
        spm += '\n'+spma.replace('$#action#$', act).replace('$#desc#$', adesc)

    dx = i.get('dict', {})

    r = ck.merge_dicts({'dict1': dx, 'dict2': dd})
    if r['return'] > 0:
        return r

    # Add entry (it will ask further questions about alias and user-friendly name)
    i['common_func'] = 'yes'
    i['dict'] = dx
    i['sort_keys'] = 'yes'
    r = ck.access(i)
    if r['return'] > 0:
        return r

    # Add module code
    p = r['path']
    pf = os.path.join(p, ck.cfg['module_full_code_name'])

    if o == 'con':
        ck.out('')
        ck.out('Creating module code '+pf+' ...')

    # Write module code
    rx = ck.save_text_file({'text_file': pf, 'string': spm})
    if rx['return'] > 0:
        return rx

    return r

##############################################################################
# show info about modules
# Notice that we now suggest to use "ck list_modules misc"


def show(i):
    """
    Input:  {
               (the same as list; can use wildcards)


            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    of = i.get('out_file', '')
    if of != '':
        xof = os.path.splitext(of)

    html = False
    if o == 'html' or i.get('web', '') == 'yes':
        html = True

    h = ''

    unique_repo = False
    if i.get('repo_uoa', '') != '':
        unique_repo = True

    import copy
    ii = copy.deepcopy(i)

    ii['out'] = ''
    ii['action'] = 'list'
    ii['add_meta'] = 'yes'

    rx = ck.access(ii)
    if rx['return'] > 0:
        return rx

    ll = sorted(rx['lst'], key=lambda k: k['data_uoa'])

    if html:
        h += 'You can install and reuse CK modules as follows:\n'
        h += '<pre>\n'
        h += ' ck pull repo:{Repo UOA - see below}\n'
        h += ' ck help {module UOA - see below}\n'
        h += '</pre>\n'

        h += 'You can check a JSON API of a given action of a given module as follows:\n'
        h += '<pre>\n'
        h += ' ck {module action - see below} {module UOA} --help\n'
        h += '</pre>\n'

        h += 'You can add your own dummy CK module as follows:\n'
        h += '<pre>\n'
        h += ' ck add module:{my module alias}\n'
        h += '</pre>\n'

        h += 'You can add a new action to the CK module as follows:\n'
        h += '<pre>\n'
        h += ' ck add_action module:{my module alias}\n'
        h += '</pre>\n'

        h += 'See <a href="https://github.com/ctuning/ck/wiki">CK documentation</a>\n'
        h += ' and the latest <a href="http://cKnowledge.org/rpi-crowd-tuning">CK paper</a> for further details.\n'

        h += '<p>\n'
        h += '<table cellpadding="4" border="1" style="border-collapse: collapse; border: 1px solid black;">\n'

        h += ' <tr>\n'
        h += '  <td nowrap><b>#</b></td>\n'
        h += '  <td nowrap><b>Module&nbsp;UOA with JSON API<br>(Python module/wrapper/plugin)</b></td>\n'
        h += '  <td nowrap><b>Repo UOA</b></td>\n'
        h += '  <td><b>Description and actions</b></td>\n'
        h += ' </tr>\n'

    repo_url = {}
    repo_private = {}

    private = ''
    num = 0
    for l in ll:
        ln = l['data_uoa']
        lr = l['repo_uoa']

        lr_uid = l['repo_uid']
        url = ''
        if lr == 'default':
            url = 'https://github.com/ctuning/ck/tree/master/ck/repo'
        elif lr_uid in repo_url:
            url = repo_url[lr_uid]
        else:
            rx = ck.load_repo_info_from_cache({'repo_uoa': lr_uid})
            if rx['return'] > 0:
                return rx
            url = rx.get('dict', {}).get('url', '')
            repo_private[lr_uid] = rx.get('dict', {}).get('private', '')
            repo_url[lr_uid] = url

        private = repo_private.get(lr_uid, '')

#        if lr not in cfg.get('skip_repos',[]) and private!='yes' and url!='':
        if lr not in cfg.get('skip_repos', []) and private != 'yes' and url != '':
            num += 1

            lm = l['meta']
            ld = lm.get('desc', '')

            actions = lm.get('actions', {})

            if lr == 'default':
                to_get = ''
            elif url.find('github.com/ctuning/') > 0:
                to_get = 'ck pull repo:'+lr
            else:
                to_get = 'ck pull repo --url='+url

            ###############################################################
            if html:
                h += ' <tr>\n'

                x1 = ''
                x2 = ''
                z1 = ''
                z11 = ''
                if url != '':
                    x1 = '<a href="'+url+'">'
                    x2 = '</a>'

                    url2 = url

                    if url2.endswith('.git'):
                        url2 = url2[:-4]

                    if '/tree/master/' not in url2:
                        url2 += '/tree/master/module/'
                    else:
                        url2 += '/module/'

                    z1 = '<a href="'+url2+ln+'/module.py">'
                    z11 = '<a href="'+url2+ln+'/.cm/meta.json">'

                h += '  <td nowrap valign="top"><a name="' + \
                    ln+'">'+str(num)+'</b></td>\n'

                h += '  <td nowrap valign="top">'+z1+ln+x2 + \
                    '</b> <i>('+z11+'CK meta'+x2+')</i></td>\n'

                h += '  <td nowrap valign="top"><b>'+x1+lr+x2+'</b></td>\n'

                h += '  <td valign="top">'+ld+'\n'

                if len(actions) > 0:
                    h += '<ul>\n'
                    for q in sorted(actions):
                        qq = actions[q]
                        qd = qq.get('desc', '')
                        h += '<li>"ck <i>'+q+'</i> '+ln+'"'
                        if qd != '':
                            h += ' - '+qd
                    h += '</ul>\n'

                h += '</td>\n'

                h += ' </tr>\n'

            ###############################################################
            elif o == 'mediawiki':
                x = lr
                if url != '':
                    x = '['+url+' '+lr+']'
                ck.out('')
                ck.out('=== '+ln+' ('+lr+') ===')
                ck.out('')
                ck.out('Desc: '+ld)
                ck.out('<br>CK Repo URL: '+x)
                if to_get != '':
                    ck.out('<br>How to get: <i>'+to_get+'</i>')
                ck.out('')
                if len(actions) > 0:

                    ck.out('Actions (functions):')
                    ck.out('')

                    for q in sorted(actions):
                        qq = actions[q]
                        qd = qq.get('desc', '')
                        ck.out('* \'\''+q+'\'\' - '+qd)

            ###############################################################
            elif o == 'con' or o == 'txt':
                if unique_repo:
                    ck.out('')
                    s = ln+' - '+ld

                else:
                    ss = ''
                    if len(ln) < 35:
                        ss = ' '*(35-len(ln))

                    ss1 = ''
                    if len(lr) < 30:
                        ss1 = ' '*(30-len(lr))

                    s = ln+ss+'  ('+lr+')'
                    if ld != '':
                        s += ss1+'  '+ld

                ck.out(s)

                if len(actions) > 0:
                    ck.out('')
                    for q in sorted(actions):
                        qq = actions[q]
                        qd = qq.get('desc', '')
                        ck.out('  * '+q+' - '+qd)

    if html:
        h += '</table>\n'

        if of != '':
            r = ck.save_text_file({'text_file': of, 'string': h})
            if r['return'] > 0:
                return r

    return {'return': 0, 'html': h}



#
# Collective Knowledge (Unified modeling using TensorFlow )
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# build model

def build(i):
    """

    Input:  {
              model_name            - model name
              (model_file)          - model output file, otherwise generated as tmp file

              features_table        - features table (in experiment module format)
              features_keys         - features flat keys 
              characteristics_table - characteristics table (in experiment module format)
              characteristics_keys  - characteristics flat keys

              (keep_temp_files)     - if 'yes', keep temp files 
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              model_input_file - temp input file (csv) - can be deleted if not keep_temp_files
              model_file       - output model file 
            }

    """

    import tempfile
    import os

    o=i.get('out','')
    oo=''
    if o=='con': oo=o

    mn=i['model_name']

    mf=i.get('model_file','')
    mf1=mf+'.tf'
    mf4=mf+'.tf.ft.txt'

    ftable=i['features_table']
    fkeys=i['features_keys']
    ctable=i['characteristics_table']
    ckeys=i['characteristics_keys']

    lftable=len(ftable)
    lctable=len(ctable)

    model_params=i.get('model_params',{})

    quiet=model_params.get('quiet','')

    # Prepare path to TF script
    xmn='module_'+mn+'.py'
    pmn=os.path.join(work['path'], xmn)
    if not os.path.isfile(pmn):
       return {'return':1, 'error': 'CK wrapper for "'+mn+'" not found ('+xmn+')'}

    # Enumerate features
    s=''

    fk=1
    for fx in sorted(fkeys): # Needed to be sorted
        s+='V'+str(fk)+') '+fx
        s+='\n'
        fk+=1

    if s!='':
       r=ck.save_text_file({'text_file':mf4, 'string':s})
       if r['return']>0: return r

    if o=='con':
       ck.out('*******************************************************')
       ck.out('Feature key convertion:')
       ck.out('')
       ck.out(s)

    if lftable!=lctable:
       return {'return':1, 'error':'length of feature table ('+str(lftable)+') is not the same as length of characteristics table ('+str(lctable)+')'}

#    if len(ckeys)>1:
#       return {'return':1, 'error':'currently we support only modeling for 1 characteristic'}

    ktf=i.get('keep_temp_files','')

    # Prepare (temporary) out model file
    fn2=mf
    if fn2=='' or i.get('web','')=='yes':
       fd2, fn2=tempfile.mkstemp(suffix='.tmp', prefix='ck-')
       os.close(fd2)
    else:
       fn2=mf1

    fn2d=fn2+'.dir'
    if os.path.isdir(fn2d):
       import shutil
       shutil.rmtree(fn2d)

    fn2+='.json'
    if os.path.isfile(fn2): os.remove(fn2)

    # Prepare temporary CSV file
    if ktf=='yes' and mf!='':
       fn1f=mf+'.train.features.csv'
       fn1c=mf+'.train.classes.csv'
       fn1j=mf+'.train.json'
    else:
       fd1f, fn1f=tempfile.mkstemp(suffix='.tmp', prefix='ck-')
       fd1c, fn1c=tempfile.mkstemp(suffix='.tmp', prefix='ck-')
       fd1j, fn1j=tempfile.mkstemp(suffix='.tmp', prefix='ck-')

       os.close(fd1f)
       os.close(fd1c)
       os.close(fd1j)

       os.remove(fn1f)
       os.remove(fn1c)
       os.remove(fn1j)

    if ktf=='yes' and o=='con': 
       ck.out('')
       ck.out('  Temporary input features file (train, CSV) = '+fn1f)
       ck.out('  Temporary input classes file (train, CSV) = '+fn1c)
       ck.out('  Temporary input file (JSON) = '+fn1j)
       ck.out('')

    ii={'action':'convert_table_to_csv',
        'module_uoa':cfg['module_deps']['experiment'],
        'table':ftable,
        'keys':fkeys,
        'file_name':fn1f,
        'csv_no_header':'yes',
        'csv_separator':',',
        'csv_decimal_mark':'.'
       }
    r=ck.access(ii)
    if r['return']>0: return r

    ii={'action':'convert_table_to_csv',
        'module_uoa':cfg['module_deps']['experiment'],
        'table':ctable,
        'keys':ckeys,
        'file_name':fn1c,
        'csv_no_header':'yes',
        'csv_separator':';',
        'csv_decimal_mark':'.'
       }
    r=ck.access(ii)
    if r['return']>0: return r

    r=ck.save_json_to_file({'json_file':fn1j, 'dict':{"ftable":ftable, 
                                                      "ctable":ctable, 
                                                      "model_params":model_params,
                                                      "output_file":fn2,
                                                      "model_dir":fn2d}})
    if r['return']>0: return r

    # Set CK environment for TF
    r=prepare_tf({'out':oo, 
                  'mode':'train', 
                  'module_name':pmn, 
                  'input_file':fn1j, 
                  'output_file':fn2,
                  'quiet':quiet})
    if r['return']>0: return r

    if ktf!='yes':
       if os.path.isfile(fn1f): os.remove(fn1f)
       if os.path.isfile(fn1c): os.remove(fn1c)
       if os.path.isfile(fn1j): os.remove(fn1j)

    if not os.path.isfile(fn2): 
       if ktf=='yes' and o=='con': 
          ck.out('')
          ck.out('  Temporary input features file (train, CSV) = '+fn1f)
          ck.out('  Temporary input classes file (train, CSV) = '+fn1c)
          ck.out('  Temporary input file (JSON) = '+fn1j)
          ck.out('')

       return {'return':1, 'error':'model was not created'}

    return {'return':0, 'model_input_file':fn1j, 'model_file':fn2}

##############################################################################
# validate model

def validate(i):
    """

    Input:  {
              model_name            - model name:
                                                  earth
                                                  lm
                                                  nnet
                                                  party
                                                  randomforest
                                                  rpart
                                                  svm

              model_file            - file with model (object) code

              features_table        - features table (in experiment module format)
              features_keys         - features flat keys 

              (keep_temp_files)     - if 'yes', keep temp files 
            }

    Output: {
              return           - return code =  0, if successful
                                             >  0, if error
              (error)          - error text if return > 0

              prediction_table - experiment table with predictions
            }

    """

    import tempfile
    import os
    import csv
    import sys

    o=i.get('out','')
    oo=''
    if o=='con': oo=o

    mn=i['model_name']

    mf=i.get('model_file','')
    mf1=mf+'.tf'

    ftable=i['features_table']
    fkeys=i.get('features_keys',[])

    ktf=i.get('keep_temp_files','')

    lftable=len(ftable)

    model_params=i.get('model_params',{})

    quiet=model_params.get('quiet','')

    # Prepare path to TF script
    xmn='module_'+mn+'.py'
    pmn=os.path.join(work['path'], xmn)
    if not os.path.isfile(pmn):
       return {'return':1, 'error': 'CK wrapper for "'+mn+'" not found ('+xmn+')'}

    # Prepare (temporary) out model file
    fn2=mf
    if fn2=='' or i.get('web','')=='yes':
       fd2, fn2=tempfile.mkstemp(suffix='.tmp', prefix='ck-')
       os.close(fd2)
    else:
       fn2=mf1

    fn2d=fn2+'.dir'
    if not os.path.isdir(fn2d):
       return {'return':1, 'error':'model directory not found ('+fn2d+')'}

    fn2+='.validate.json'
    if os.path.isfile(fn2): os.remove(fn2)

    if ktf=='yes':
       fn1f=mf+'.test.features.csv'
       fn1j=mf+'.test.json'
    else:
       fd1f, fn1f=tempfile.mkstemp(suffix='.tmp', prefix='ck-')
       fd1j, fn1j=tempfile.mkstemp(suffix='.tmp', prefix='ck-')

       os.close(fd1f)
       os.close(fd1j)

       os.remove(fn1f)
       os.remove(fn1j)

    if ktf=='yes' and o=='con': 
       ck.out('')
       ck.out('  Temporary input features file (test, CSV) = '+fn1f)
       ck.out('  Temporary input file (JSON) = '+fn1j)
       ck.out('')

    ii={'action':'convert_table_to_csv',
        'module_uoa':cfg['module_deps']['experiment'],
        'table':ftable,
        'keys':fkeys,
        'file_name':fn1f,
        'csv_no_header':'yes',
        'csv_separator':';',
        'csv_decimal_mark':'.'
       }
    r=ck.access(ii)
    if r['return']>0: return r

    r=ck.save_json_to_file({'json_file':fn1j, 'dict':{"ftable":ftable, 
                                                      "output_file":fn2,
                                                      "model_dir":fn2d}})
    if r['return']>0: return r

    # Set CK environment for TF
    r=prepare_tf({'out':oo, 
                  'mode':'prediction', 
                  'module_name':pmn, 
                  'input_file':fn1j, 
                  'output_file':fn2,
                  'quiet':quiet})
    if r['return']>0: return r

    if ktf!='yes':
       if os.path.isfile(fn1f): os.remove(fn1f)
       if os.path.isfile(fn1j): os.remove(fn1j)

    if not os.path.isfile(fn2): 
       if ktf=='yes' and o=='con': 
          ck.out('')
          ck.out('  Temporary input features file (test, CSV) = '+fn1f)
          ck.out('  Temporary input file (JSON) = '+fn1j)
          ck.out('')

       return {'return':1, 'error':'model was not created'}

    # Read output file
    r=ck.load_json_file({'json_file':fn2})
    if r['return']>0: return r

    pr=r['dict']['ctable']

    if ktf!='yes': os.remove(fn2)

    pr1=[]
    for q in pr:
        pr1.append([q])

    return {'return':0, 'prediction_table':pr1}

##############################################################################
# Convert categorical values to floats

def prepare_tf(i):
    """
    Input:  {
              (tags)      - extra tags to select tensorflow

              module_name - CK wrapper for specific TF model
              mode        - train or predict
              input_file  - input file

              (quiet)     - if 'yes', select first available TF
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

            }

    """

    import locale

    o=i.get('out','')
    oo=''
    if o=='con': 
       oo=o
       ck.out('')
       ck.out('Detecting TensorFlow installations via CK ...')
       ck.out('')

    # Check if ck-tensorflow repo is there
    r=ck.access({'action':'load',
                 'module_uoa':cfg['module_deps']['soft'],
                 'data_uoa':cfg['data_deps']['soft_lib_tensorflow']})
    if r['return']>0:
       if r['return']!=16: return r
       return {'return':1, 'error':'ck-tensorflow repo is not installed. Please installed it using "ck pull repo:ck-tensorflow" and try again'}

    module_name=i['module_name']
    mode=i['mode']
    fi=i['input_file']

    # Prepare TF CK environment
    tags='lib,tensorflow'
    if i.get('tags','')!='': tags+=','+tags

    r=ck.access({'action':'set',
                 'module_uoa':cfg['module_deps']['env'],
                 'tags':tags,
                 'version_from':[1,4,0],
                 'quiet':i.get('quiet',''),
                 'out':oo})
    if r['return']>0: return r

    sb=r['bat']
    d=r['dict']

    # Check python
    pf=d.get('deps',{}).get('python',{}).get('dict',{}).get('env',{}).get('CK_ENV_COMPILER_PYTHON_FILE','')
    if pf=='':
       return {'return':1, 'error':'can\'t find associated python in the selected TF (maybe installed without python?)'}

    sb+='\n'
    sb+=pf+' '+module_name+' '+mode+' '+fi
    sb+='\n'

    if oo=='con':
       ck.out('')
       ck.out('  Executing command:')
       ck.out('')
       ck.out(sb)
       ck.out('')

    r=ck.access({'action':'shell',
                 'module_uoa':cfg['module_deps']['os'],
                 'encoding':locale.getdefaultlocale()[1],
                 'output_to_console':'yes',
                 'cmd':sb})
    if r['return']>0: return r
  
    return {'return':0, 'bat':sb, 'python_file':pf}



#
# Collective Knowledge (platform detection)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
sep='***************************************************************************************'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# collect info about platforms

def detect(i):
    """
    Input:  {
              (target)               - if specified, use info from 'machine' module

              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)
              (skip_device_init)     - if 'yes', do not initialize device
              (print_device_info)    - if 'yes', print extra device info

              (skip_info_collection) - if 'yes', do not collect info (particularly for remote)

              (exchange)             - if 'yes', exchange info with some repo (by default, remote-ck)
              (share)                - the same as 'exchange'
              (exchange_repo)        - which repo to record/update info (remote-ck by default)
              (exchange_subrepo)     - if remote, remote repo UOA
              (exchange_locally)     - if 'yes', exchange locally

              (force_platform_name)  - if !='', use this for platform name

              (extra_info)           - extra info about author, etc (see add from CK kernel)

              (quiet)                - do not ask questions whenever possible
              (skip_gpu_info)        - if 'yes', do not collect GPU info
              (platform_init_uoa)    - if !='', use these platform.init scripts
              (update_platform_init) - update platform.init scripts (ask user)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              host_os_uoa                 - host OS UOA
              host_os_uid                 - host OS UID
              host_os_dict                - host OS meta

              os_uoa                      - target OS UOA
              os_uid                      - target OS UID
              os_dict                     - target OS meta

              (devices)                   - return devices if device_id==''
              (device_id)                 - if device_id=='' and only 1 device, select it

              features = {
                cpu            - CPU features (properties), unified
                cpu_misc       - assorted CPU features (properties), platform dependent

                os             - OS features (properties), unified
                os_misc        - assorted OS features (properties), platform dependent

                platform       - platform features (properties), unified
                platform_misc  - assorted platform features (properties), platform dependent

                gpu            - GPU features (properties), unified
                gpu_misc       - assorted GPU features (properties), platform dependent
              }
            }

    """

    import os

    o=i.get('out','')

    oo=''
    if o=='con': oo=o

    # Check if target
    if i.get('target','')!='':
       r=ck.access({'action':'init',
                    'module_uoa':cfg['module_deps']['machine'],
                    'input':i})
       if r['return']>0: return r

    device_cfg=i.get('device_cfg',{})

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    sic=i.get('skip_info_collection','')
    sdi=i.get('skip_device_init','')
    pdv=i.get('print_device_info','')
    ex=i.get('exchange','')
    if ex=='': ex=i.get('share','')

    quiet=i.get('quiet','')
    sgi=i.get('skip_gpu_info','')
    piuoa=i.get('platform_init_uoa','')

    einf=i.get('extra_info','')
    if einf=='': einf={}

    # If exchange, check that repo from this env is cached and recache if needed
    el=i.get('exchange_locally','')
    if el!='yes' and ex=='yes':
       er=i.get('exchange_repo','')

       if er!='':
          rx=ck.load_repo_info_from_cache({'repo_uoa':er})
          if rx['return']>0: 
             if o=='con':
                ck.out('')
                ck.out('It appears that CK remote repo ('+er+') is not in CK cache - recaching ...')

             rx=ck.access({'action':'recache',
                           'module_uoa':cfg['module_deps']['repo']})
             if rx['return']>0: return rx

    # Get OS info ###############################################################
    if oo=='con': 
       ck.out(sep)
       ck.out('Detecting OS and CPU features ...')

    import copy
    ii=copy.deepcopy(i)
    ii['out']=oo
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.cpu']
    rr=ck.access(ii) # DO NOT USE rr further - will be reused as return !
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tosd.update(device_cfg.get('update_target_os_dict',{}))

    tbits=tosd.get('bits','')

    tdid=rr['device_id']

    # Some params
    ro=tosd.get('redirect_stdout','')
    win=tosd.get('windows_base','')

    dv=''
    if tdid!='': dv=' -s '+tdid

    spd=tosd.get('skip_platform_detection','')
    if spd=='yes':
       sgi='yes'

    # Init
    prop={}
    prop_all={}

    xos=rr['os_uoa']
    device_id=rr['device_id']

    os_uoa=rr['os_uoa']
    os_uid=rr['os_uid']
    os_dict=rr['os_dict']

    remote=os_dict.get('remote','')
    remote_ssh=os_dict.get('remote_ssh','')
    os_win=os_dict.get('windows_base','')
    os_mac=os_dict.get('macos','')

    ro=os_dict.get('redirect_stdout','')

    # Get GPU info ####################################################
    if sgi!='yes':
       if oo=='con': 
          ck.out(sep)
          ck.out('Detecting GPU features ...')

       import copy
       ii=copy.deepcopy(i)
       ii['out']=oo
       ii['skip_print_os_info']='yes'
       ii['action']='detect'
       ii['module_uoa']=cfg['module_deps']['platform.gpu']
       rx=ck.access(ii) # DO NOT USE rr further - will be reused as return !
       if rx['return']>0: return rx

       # Merge with other features
       ry=ck.merge_dicts({'dict1':rr, 'dict2':rx})
       if ry['return']>0: return ry
       rr=ry['dict1']

    # Get info about system ######################################################
    if oo=='con': 
       ck.out(sep)
       ck.out('Detecting system features ...')

    if remote=='yes' and remote_ssh!='yes':
       params={}

       rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
       if rx['return']>0: return rx
       fn=rx['file_name']

       x=tosd.get('adb_all_params','')
       x=x.replace('$#redirect_stdout#$', ro)
       x=x.replace('$#output_file#$', fn)

       dv=''
       if tdid!='': dv=' -s '+tdid
       x=x.replace('$#device#$',dv)

       if o=='con' and pdv=='yes':
          ck.out('')
          ck.out('Receiving all parameters:')
          ck.out('  '+x)

       rx=os.system(x)
       if rx!=0:
          if o=='con':
             ck.out('')
             ck.out('Non-zero return code :'+str(rx)+' - likely failed')
          return {'return':1, 'error':'access to remote device failed'}

       # Read and parse file
       rx=ck.load_text_file({'text_file':fn, 'split_to_list':'yes', 'delete_after_read':'yes'})
       if rx['return']>0: return rx
       ll=rx['lst']

       for s in ll:
           s1=s.strip()

           q2=s1.find(']: [')
           k=''
           if q2>=0:
              k=s1[1:q2].strip()
              v=s1[q2+4:].strip()
              v=v[:-1].strip()

              params[k]=v

       prop_all['adb_params']=params

#       for q in params:
#           v=params[q]
#           print q+'='+v

       model=params.get('ro.product.model','')
       manu=params.get('ro.product.manufacturer','')
       if model!='' and manu!='':
          if model.lower().startswith(manu.lower()):
             model=model[len(manu)+1:]

       if manu=='' and model!='': manu=model

       manu=manu.upper()
       model=model.upper()

       prop['name']=manu
       if model!='': prop['name']+=' '+model
       prop['model']=model
       prop['vendor']=manu
    elif spd!='yes':
       x1=''
       x2=''

       target_system_model=''
       target_name=''

       if os_win=='yes':
          r=get_from_wmic({'group':'csproduct',
                           'remote_shell':tosd.get('remote_shell','').replace('$#device#$',dv),
                           'remote_shell_end':tosd.get('remote_shell_end','')})
          if r['return']>0: return r
          info1=r['dict']

          x1=info1.get('Vendor','')
          x2=info1.get('Version','')

          target_name=x1+' '+x2

          r=get_from_wmic({'cmd':'computersystem get model',
                           'remote_shell':tosd.get('remote_shell','').replace('$#device#$',dv),
                           'remote_shell_end':tosd.get('remote_shell_end','')})
          if r['return']>0: return r
          target_system_model=r['value']

          prop_all['cs_product']=info1
       elif os_mac=='yes':
          ################################################### Obtaining info on MacOS (including remote)
          cmd=['system_profiler', 'SPHardwareDataType']

          if remote_ssh=='yes':
             cmd=tosd.get('remote_shell','').replace('$#device#$',dv)+' system_profiler SPHardwareDataType '+tosd.get('remote_shell_end','')

          r=ck.access({'action':'run_and_get_stdout',
                       'module_uoa':cfg['module_deps']['os'],
                       'cmd': cmd})
          if r['return']>0: return r

          x1='Apple Inc.'

          for line in r['stdout'].splitlines():
            if ':' in line:
              left, right = line.split(':', 1)
              left = left.strip().lower()
              if left == 'model name':
                target_name = right.strip()
              if left == 'model identifier':
                target_system_model = right.strip()
              if target_name!='' and target_system_model!='':
                break

       else:
          ################################################### Obtaining info on Linux (including remote)
          cmd=['cat', '/sys/devices/virtual/dmi/id/sys_vendor']

          if remote_ssh=='yes':
             cmd=tosd.get('remote_shell','').replace('$#device#$',dv)+' cat /sys/devices/virtual/dmi/id/sys_vendor '+tosd.get('remote_shell_end','')

          r=ck.access({'action':'run_and_get_stdout',
                       'module_uoa':cfg['module_deps']['os'],
                       'cmd': cmd})
          if r['return']>0: return r
          if r['stdout'].strip()!='': x1=r['stdout'].strip()

          cmd=['cat', '/sys/devices/virtual/dmi/id/product_version']

          if remote_ssh=='yes':
             cmd=tosd.get('remote_shell','').replace('$#device#$',dv)+' cat /sys/devices/virtual/dmi/id/product_version '+tosd.get('remote_shell_end','')

          r=ck.access({'action':'run_and_get_stdout',
                       'module_uoa':cfg['module_deps']['os'],
                       'cmd': cmd})
          if r['return']>0: return r
          if r['stdout'].strip()!='': x2=r['stdout'].strip()

          if x1!='' and x2!='':
             target_name=x1+' '+x2

          cmd=['cat', '/sys/devices/virtual/dmi/id/product_name']

          if remote_ssh=='yes':
             cmd=tosd.get('remote_shell','').replace('$#device#$',dv)+' cat /sys/devices/virtual/dmi/id/product_name '+tosd.get('remote_shell_end','')

          r=ck.access({'action':'run_and_get_stdout',
                       'module_uoa':cfg['module_deps']['os'],
                       'cmd': cmd})
          if r['return']>0: return r
          if r['stdout'].strip()!='': target_system_model=r['stdout'].strip()

          #############################################################################################################
          # If empty, try platform specific probes (for now here and later maybe in separate platform-specific scripts)
          if target_name=='' and target_system_model=='':
             # Check RPi
             fail=False
             try:
                import RPi.GPIO as GPIO
             except Exception as e:
                fail=True

             if not fail:
                try:
                   rpi=GPIO.RPI_INFO
                   x1=rpi.get('MANUFACTURER','')
                   target_name='Raspberry Pi'
                   target_system_model='Raspberry '+rpi.get('TYPE','')
                except Exception as e: 
                   pass

          if target_name=='' and target_system_model=='':
             # Check via /proc/device-tree/model
             rx=ck.load_text_file({'text_file':'/proc/device-tree/model'})
             if rx['return']==0:
                target_system_model=rx['string'].strip()
                zz1=target_system_model.split(' ')
                if len(zz1)>0:
                   x1=zz1[0]

       prop['vendor']=x1
       if target_name=='' and x1!='': target_name=x1
       prop['name']=target_name
       if target_system_model!='': prop['name']+=' ('+target_system_model+')'
       prop['model']=target_system_model

       fpn=i.get('force_platform_name','')
       if fpn!='':
          prop['name']=fpn

    if o=='con':
       ck.out('')
       ck.out('Platform name:   '+prop.get('name',''))
       ck.out('Platform vendor: '+prop.get('vendor',''))
       ck.out('Platform model:  '+prop.get('model',''))

    fuoa=''
    fuid=''

    # Exchanging info #################################################################
    if ex=='yes':
       er=i.get('exchange_repo','')
       esr=i.get('exchange_subrepo','')
       el=i.get('exchange_locally','')

       if el!='yes' and er=='': 
          er=ck.cfg['default_exchange_repo_uoa']
          esr=ck.cfg['default_exchange_subrepo_uoa']

       if o=='con':
          ck.out('')
          ck.out('Exchanging information with '+er+' repository ...')

       xn=prop.get('name','')
       if xn=='':
          # Check if exists in configuration

          dcfg={}
          ii={'action':'load',
              'module_uoa':cfg['module_deps']['cfg'],
              'data_uoa':cfg['cfg_uoa']}
          r=ck.access(ii)
          if r['return']>0 and r['return']!=16: return r
          if r['return']!=16:
             dcfg=r['dict']

          dx=dcfg.get('platform_name',{}).get(tos,{})
          x=tdid
          if x=='': x='default'
          xn=dx.get(x,'')

          if (xn=='' and o=='con'):
             r=ck.inp({'text':'Enter your platform name (for example Samsung Chromebook 2, Huawei Ascend Mate 7, IBM SyNAPSE): '})
             xxn=r['string'].strip()

             if xxn!=xn:
                xn=xxn

                if 'platform_name' not in dcfg: dcfg['platform_name']={}
                if tos not in dcfg['platform_name']: dcfg['platform_name'][tos]={}
                dcfg['platform_name'][tos][x]=xn

                ii={'action':'update',
                    'module_uoa':cfg['module_deps']['cfg'],
                    'data_uoa':cfg['cfg_uoa'],
                    'dict':dcfg}
                r=ck.access(ii)
                if r['return']>0: return r

          if xn=='':
             return {'return':1, 'error':'can\'t exchange information where main name is empty'}

          ixn=xn.find(' ')
          if ixn>0: 
             xx=xn[:ixn].strip()
             prop['vendor']=xx
             prop['model']=xn[ixn+1:].strip()

          prop['name']=xn

       ii={'action':'exchange',
           'module_uoa':work['self_module_uid'],
           'sub_module_uoa':work['self_module_uid'],
           'repo_uoa':er,
           'data_name':prop.get('name',''),
           'extra_info':einf,
           'all':'yes',
           'dict':{'features':prop}} # Later we should add more properties from prop_all,
                                     # but should be careful to remove any user-specific info
       if esr!='': ii['remote_repo_uoa']=esr
       r=ck.access(ii)
       if r['return']>0: return r

       fuoa=r.get('data_uoa','')
       fuid=r.get('data_uid','')

       prop=r['dict'].get('features',{})

       if o=='con' and r.get('found','')=='yes':
          ck.out('  Platform CK entry already exists ('+fuid+') - loading latest meta (features) ...')

    if 'features' not in rr: rr['features']={}

    rr['features']['platform']=prop
    rr['features']['platform_misc']=prop_all

    if fuoa!='' or fuid!='':
       rr['features']['platform_uoa']=fuoa
       rr['features']['platform_uid']=fuid

    return rr

##############################################################################
# Get info from WMIC on Windows

def get_from_wmic(i):
    """
    Input:  {
              cmd     - cmd for wmic
              (group) - get the whole group

              (remote_shell)     - remote shell prefix if needed
              (remote_shell_end) - remote shell suffix if needed
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              value        - obtained value
              (dict)       - if group
            }

    """

    import os

    value=''
    dd={}

    rs=i.get('remote_shell','')
    rse=i.get('remote_shell_end','')

    xcmd=i.get('cmd','')
    xgroup=i.get('group','')
    if xgroup!='': xcmd=xgroup

    # We need to use file since it is encoded with UTF-16 so will need to load it with this encoding
    rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
    if rx['return']>0: return rx
    fn=rx['file_name']

    cmd=rs+' wmic '+xcmd+' '+rse+' > '+fn

    r=os.system(cmd)
    if r!=0:
       return {'return':1, 'error':'command returned non-zero value: '+cmd}

    # Read and parse file
    enc='utf16'
    if rs!='': enc=''

    rx=ck.load_text_file({'text_file':fn, 'encoding':enc, 'split_to_list':'yes'})
    if rx['return']>0: return rx
    ll=rx['lst']

    if os.path.isfile(fn): os.remove(fn)

    if xgroup=='':
       if len(ll)>1:
          value=ll[1].strip()
    else:
       if len(ll)>1:
          kk=ll[0]
          value=ll[1]

          xkeys=kk.split(' ')
          keys=[]
          for q in xkeys:
              if q!='': keys.append(q)

          for q in range(0, len(keys)):
              k=keys[q]

              if q==0: qx=0
              else: 
                 y=' '
                 if q==len(keys)-1: y=''
                 qx=kk.find(' '+k+y)

              if q==len(keys)-1:
                 qe=len(value)
              else:
                 qe=kk.find(' '+keys[q+1]+' ')

              v=value[qx:qe].strip()
              dd[k]=v

    return {'return':0, 'value':value, 'dict':dd}

##############################################################################
# Init remote device

def init_device(i):
    """
    Input:  {
              os_dict      - OS dict to get info about how to init device
              device_id    - ID of the device if more than one
              (key)        - key from OS to use, by default remote_init
                             useful for deinitialization, i.e. use key=remote_deinit
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o=i.get('out','')

    key=i.get('key','')
    if key=='': key='remote_init'

    osd=i['os_dict']
    tdid=i['device_id'].strip()

    ri=osd.get(key,'')
    if ri!='':
       dv=''
       if tdid!='': dv=' -s '+tdid
       ri=ri.replace('$#device#$',dv)

       if o=='con':
          ck.out('Initializing remote device:')
          ck.out('')
          ck.out('  '+ri)

       rx=os.system(ri)
       if rx!=0:
          if o=='con':
             ck.out('')
             ck.out('Non-zero return code :'+str(rx)+' - likely failed')
          return {'return':1, 'error':'remote device initialization failed'}

       device_init='yes'

    return {'return':0}

##############################################################################
# exchange properties

def exchange(i):
    """
    Input:  {
              sub_module_uoa - module to search/exchange
              data_name      - data name to search

              (search_dict)  - search entry by this meta

              (repo_uoa)     - where to record

              (dict)         - dictionary to check/record

              (all)          - if 'yes', check all dict['features'] and add to separate file 

              (extra_info)   - extra info about author, etc (see add from CK kernel)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              dict         - if exists, load updated dict (can be collaboratively extended to add more properties 
                                        (or unique/representative species -> software, hardware, gpu, programs, data sets!)
              (found)      - if 'yes', entry was found
            }

    """

    import os

    ruoa=i.get('repo_uoa','')
    smuoa=i['sub_module_uoa']

    dname=i.get('data_name','')

    dd=i.get('dict',{})
    ddf=dd.get('features',{})

    sd=i.get('search_dict',{})

    al=i.get('all','')

    if dname=='' and len(sd)==0:
       return {'return':1, 'error':'name and search meta are empty in platform information exchange'}

    # Search if already exists (and not only in upload)
    ii={'action':'search',
        'module_uoa':smuoa,
# FGG: I commented next line since we can move 
#      well-known entries to other repositories
#      such as ck-crowdtuning instead of upload
#       'repo_uoa':ruoa,
        'ignore_case':'yes'}
    if dname!='':
       ii['search_by_name']=dname
    if len(sd)>0:
       ii['search_dict']=sd

    rx=ck.access(ii)
    if rx['return']>0: return rx
    lst=rx['lst']

    if len(lst)==0:
       ei=i.get('extra_info',{})

       # Add info
       rx=ck.access({'action':'add',
                     'module_uoa':smuoa,
                     'repo_uoa':ruoa,
                     'data_name':dname,
                     'dict':dd,
                     'extra_info':ei,
                     'sort_keys':'yes'})

    else:
       # Load
       ll=lst[0]
       duoa=ll.get('data_uid','')
       xruoa=ll.get('repo_uoa','')
       rx=ck.access({'action':'load',
                     'module_uoa':smuoa,
                     'repo_uoa':xruoa,
                     'data_uoa':duoa})

       rx['found']='yes'

    if rx['return']>0: return rx

    if al=='yes':
       # We also record all info only if forced to do so
       #  by CK local kernel configuration. It is needed to aggregate
       #  such info on experiment crowdsourcing servers
       #  and then share this info via GitHub,
       #  but not locally, otherwise it will be a mess...

       rap=ck.cfg.get('record_all_platform_info','')
#          if rap=='yes':
#
#             # Not parallel usage safe (on the other hand, will not loose too much at the moment) ...
#
#             # Check if extra parameters are saved
#             p=rx['path']
#             p1=os.path.join(p, 'all.json')
#
#             d={'all':[]}
#             toadd=True
#
#             touched=0
#
#             if os.path.isfile(p1):
#                ry=ck.load_json_file({'json_file':p1})
#                if ry['return']>0: return ry
#                d=ry['dict']
#
#                touched=d.get('touched',0)
#                touched+=1
#
#                if 'all' not in d: d['all']=[]
#                dall=d.get('all',[])
#
#                for q in dall:
#                    rz=ck.compare_dicts({'dict1':q, 'dict2':ddf})
#                    if rz['return']>0: return rz
#                    if rz['equal']=='yes':
#                       toadd=False
#                       break
#
#             d['touched']=touched
#
#             if toadd:
#                d['all'].append(ddf)
#
#             rz=ck.save_json_to_file({'json_file':p1, 'dict':d})
#             if rz['return']>0: return rz

    return rx

##############################################################################
# deinitialize device (put to powersave mode)

def deinit(i):
    """
    Input:  {
              (target)               - target machine if needed (to correctly handle remote execution)

              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)
              (device_id)            - device id if remote (such as adb)
              (key)                  - {'remote_init' or 'remote_deinit'}
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    # Check if target
    if i.get('target','')!='':
       r=ck.access({'action':'init',
                    'module_uoa':cfg['module_deps']['machine'],
                    'input':i})
       if r['return']>0: return r

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    # Get OS info
    import copy
    ii=copy.deepcopy(i)
    ii['out']=o
    ii['action']='detect'
    ii['skip_device_init']='yes'     # consider that device was already initialized
    ii['skip_info_collection']='yes'
    ii['module_uoa']=cfg['module_deps']['platform.os']
    rr=ck.access(ii) # DO NOT USE rr further - will be reused as return !
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tbits=tosd.get('bits','')

    tdid=rr['device_id']

    key=i.get('key','')
    if key=='': key='remote_deinit'

    ii={'os_dict':tosd,
        'device_id':tdid,
        'key':key,
        'out':o}
    return init_device(ii)

##############################################################################
# viewing entries as html

def show(i):
    """
    Input:  {
              data_uoa
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html         - generated HTML
            }

    """


    h='<h2>Platforms participating in crowd-tuning</h2>\n'

    h+='<i>Reuse/extend <a href="https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform">CK JSON meta information</a> of these platforms using "ck pull repo:ck-crowdtuning-platforms" ...</i><br><br>\n'

    h+='<table class="ck_table" border="0" cellpadding="6" cellspacing="0">\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    h+=' <tr style="background-color:#cfcfff;">\n'
    h+='  <td><b>\n'
    h+='   #\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Vendor\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Model\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':">CK UID</a>\n'
    h+='  </b></td>\n'
    h+=' </tr>\n'

    ruoa=i.get('repo_uoa','')
    muoa=work['self_module_uoa']
    duoa=i.get('data_uoa','')

    r=ck.access({'action':'search',
                 'module_uoa':muoa,
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'add_info':'yes',
                 'add_meta':'yes'})
    if r['return']>0: 
       return {'return':0, 'html':'Error: '+r['error']}

    lst=r['lst']

    num=0
    for q in sorted(lst, key = lambda x: (x.get('meta',{}).get('features',{}).get('vendor','').upper(), \
                                          x.get('meta',{}).get('features',{}).get('model','').upper())):

        num+=1

        duoa=q['data_uoa']
        duid=q['data_uid']

        meta=q['meta']
        ft=meta.get('features',{})

        name=ft.get('name','')
        vendor=ft.get('vendor','')
        model=ft.get('model','')

        h+=' <tr>\n'
        h+='  <td valign="top">\n'
        h+='   '+str(num)+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+vendor+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+model+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':'+duid+'">'+duid+'</a>\n'
        h+='  </td>\n'
        h+=' </tr>\n'


    h+='</table><br><br>\n'

    return {'return':0, 'html':h}

##############################################################################
# browse platforms participated in experiment crowdsourcing (crowd-benchmarking and crowd-tuning)

def browse(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # TBD: should calculate url
    url='https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform'

    import webbrowser
    webbrowser.open(url)

    import time
    time.sleep(3)

    url='http://cknowledge.org/repo/web.php?action=index&module_uoa=wfe&native_action=show&native_module_uoa=platform'

    import webbrowser
    webbrowser.open(url)

    return {'return':0}



#
# Collective Knowledge (dataset features)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# extract dataset features

def extract(i):
    """
    Input:  {
              (repo_uoa)        - repository UOA
              (data_uoa)        - dataset UOA (can be wildcards)

              (tags)            - tags to process specific datasets

              (target_repo_uoa) - repo, where to save features - if =='', use repo_uoa


            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              dict         - final dict with key 'features'={...} 
            }

    """

    import os
    import json

    o=i.get('out','')

    muoa=cfg['module_deps']['dataset']
    duoa=i.get('data_uoa','')
    ruoa=i.get('repo_uoa','')

    truoa=i.get('target_repo_uoa','')
    if truoa=='' and ruoa!='': truoa=ruoa

    tags=i.get('tags','')

    rx=ck.access({'action':'search',
                  'repo_uoa':ruoa,
                  'module_uoa':muoa,
                  'data_uoa':duoa,
                  'tags':tags})
    if rx['return']>0: return rx

    lst=rx['lst']

    feat1={}

    for q in lst:
        duid=q['data_uid']
        duoa=q['data_uoa']

        if o=='con':
           ck.out('Processing '+duoa+' ...')

        rx=ck.access({'action':'load',
                      'module_uoa':muoa,
                      'data_uoa':duoa})
        if rx['return']>0: return rx

        d=rx['dict']
        p=rx['path']

        df=d.get('dataset_files','')
        dt=d.get('tags','')

        feat={}

        otags=d.get('tags',[])

        ddd={'tags':otags, 'dataset_uid':duid, 'dataset_uoa':duoa}

        ts=0
        for f in df:
            p1=os.path.join(p,f)
            if os.path.isfile(p1):
               ts+=os.path.getsize(p1)

        if ts!=0: feat['total_size']=ts

        if 'image' in dt:
           if o=='con':
              ck.out('  Image detected.')

              for f in df:
                  p1=os.path.join(p,f)
                  if os.path.isfile(p1):
                     try:
                       from PIL import Image
                       im = Image.open(p1)

                       feat['mode']=str(im.mode)
                       feat['format']=str(im.format)
                       feat['width']=im.size[0]
                       feat['height']=im.size[1]

                       inf=im.info
                       feat['compression']=inf.get('compression','')
                       dpi=inf.get('dpi',[])
                       if len(dpi)>1:
                          feat['xdpi']=dpi[0]
                          feat['ydpi']=dpi[1]
                       
                       feat['raw_info']=im.info
                     except Exception as e: 
                        pass

        if len(feat)>0:
           rr=ck.dumps_json({'dict':feat, 'sort_keys':'yes', 'skip_indent':'yes'})
           if rr['return']>0: 
              if 'raw_info' in feat: # Usually source of problems
                 del(feat['raw_info'])
                 rr=ck.dumps_json({'dict':feat, 'sort_keys':'yes', 'skip_indent':'yes'})
                 if rr['return']>0: return rr
              else:
                 return rr

           sfeat=rr['string']
           ck.out('  '+sfeat)

           found=False
           ry=ck.access({'action':'load',
                         'module_uoa':work['self_module_uid'],
                         'data_uoa':duid})
           if ry['return']==0: 
              ddd=ry['dict']              
              found=True

           feat1=ddd.get('features',{})
           rz=ck.merge_dicts({'dict1':feat1, 'dict2':feat})
           if rz['return']>0: return rz
           feat1=rz['dict1']

           ddd['features']=feat1
           ddd['tags']=otags

           ii={}
           ii['action']='add'
           if found: ii['action']='update'
           ii['module_uoa']=work['self_module_uid']
           ii['data_uoa']=duoa
           ii['data_uid']=duid
           ii['repo_uoa']=truoa
           ii['dict']=ddd
           ii['substitute']='yes'
           ry=ck.access(ii)
           if ry['return']>0: return ry

    return {'return':0, 'dict':{'features':feat1}}

##############################################################################
# converting raw RGB image to png or other formats

def convert_raw_rgb_image(i):
    """
    Input:  {
              input_file    - input raw RGB file
              output_file   - output file
              width         - image width
              height        - image height
              (mode)        - mode: RGB (default), RGBA, ...
              (output_type) - type of output file: PNG (default), JPEG ...
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    fi=i['input_file']
    fo=i['output_file']

    w=int(i['width'])
    h=int(i['height'])

    ot=i.get('output_type','')
    if ot=='': ot='png'

    mode=i.get('mode','')
    if mode=='': mode='RGB'

    # Load binary file
    r=ck.load_text_file({'text_file':fi, 'keep_as_bin':'yes'})
    if r['return']>0: return r

    bin=r['bin']

    # Create image
    from PIL import Image

    im = Image.frombuffer(mode, (w,h), bin, "raw", mode, 0, 1)

    im.save(fo, ot)

    return {'return':0}



#
# Collective Knowledge (platform - GPGPU)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Detect GPGPU

def detect(i):
    """
    Input:  {
              (target)               - if specified, use info from 'machine' module

              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)
              (skip_device_init)     - if 'yes', do not initialize device
              (print_device_info)    - if 'yes', print extra device info

              (skip_info_collection) - if 'yes', do not collect info (particularly for remote)

              (skip_print_os_info)   - if 'yes', do not print OS info

              (exchange)             - if 'yes', exchange info with some repo (by default, remote-ck)
              (share)                - the same as 'exchange'
              (exchange_repo)        - which repo to record/update info (remote-ck by default)
              (exchange_subrepo)     - if remote, remote repo UOA
              (exchange_locally)     - if 'yes', exchange locally

              (extra_info)           - extra info about author, etc (see add from CK kernel)

              (type)                 - cuda or opencl
              (quiet)                - select default dependencies

              (select)               - if 'yes', select specific platform and device (only when one type, i.e. opencl or cuda)
              (deps)                 - pre-set dependencies when used inside 'program' module (to reuse already selected compilers and libs)

              (compute_platform_id)  - pre-select platform ID
              (compute_device_id)    - pre-select device ID

              (sudo)                 - force SUDO use ...
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              features = [
                {
                  gpgpu          - GPGPU features (properties), unified
                  gpgpu_misc     - assorted GPGPU features (properties), platform dependent
                  gpgpu_id       - local ID {'gpgpu_platform_id', 'gpgpu_device_id'}
                }
              ]

              choices = {
                  (compute_platform_id)
                  (compute_device_id)
              }
            }

    """

    import os
    import copy

    o=i.get('out','')

    oo=''
    if o=='con': oo=o

    try:
        pcur=os.getcwd()
    except OSError:
        os.chdir('..')
        pcur=os.getcwd()

    quiet=i.get('quiet','')

    # Check if target
    if i.get('target','')!='':
       r=ck.access({'action':'init',
                    'module_uoa':cfg['module_deps']['machine'],
                    'input':i})
       if r['return']>0: return r

    device_cfg=i.get('device_cfg',{})

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    sic=i.get('skip_info_collection','')
    sdi=i.get('skip_device_init','')
    pdv=i.get('print_device_info','')

    ex=i.get('exchange','')
    if ex=='': ex=i.get('share','')

    einf=i.get('extra_info','')
    if einf=='': einf={}

    # Get OS info
    ii=copy.deepcopy(i)
    ii['out']=oo
    if i.get('skip_print_os_info','')=='yes': ii['out']=''
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.cpu']
    rr=ck.access(ii) # DO NOT USE rr further - will be reused as return !
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tosd.update(device_cfg.get('update_target_os_dict',{}))

    tbits=tosd.get('bits','')

    tdid=rr['device_id']

    remote=tosd.get('remote','')

    # Some params
    ro=tosd.get('redirect_stdout','')
    remote=tosd.get('remote','')
    win=tosd.get('windows_base','')

    stdirs=tosd.get('dir_sep','')

    dv=''
    if tdid!='': dv=' -s '+tdid

    props=[]

    fuid=''
    fuoa=''

    # Check if program to get CUDA device exists
    types=['cuda','opencl']
    tp=i.get('type','')

    if tp=='':
       if i.get('opencl','')=='yes': tp='opencl'
       elif i.get('cuda','')=='yes': tp='cuda'

    tpo=tp # to quit later if a specific type needs to be detected but env doesn't exist in the CK

    if tp!='': types=[tp]

    for tp in types:
        prop={}
        prop_id={}
        prop_all={}
        deps={}

        if o=='con':
           ck.out('************************************************')
           ck.out('Detecting GPGPU type: '+tp)
           ck.out('')

        tags=cfg['tool_tags'][tp]

        # Get tool env
        ii={'action':'set',
            'module_uoa':cfg['module_deps']['env'],
            'host_os':hos,
            'target_os':tos,
            'device_id':tdid,
            "skip_default": "yes",
            'tags':tags,
            'out':oo}
        r=ck.access(ii)
        if r['return']>0: 
           # Fail only if a given type (CUDA/OpenCL) is specified, otherwise continue!
           if tpo=='': continue
           return r

        env=r['env']
        deps=r['dict']['deps']

        # Try to run program
        rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.tmp'})
        if rx['return']>0: return rx
        ftmp=rx['file_name']

        pf=env[cfg['tool_env_key'][tp]]

        if remote=='yes':
            dv=''
            if tdid!='': dv=' -s '+tdid

            remote_init=tosd.get('remote_init','').replace('$#device#$',dv)
            remote_push=tosd.get('remote_push','').replace('$#device#$',dv)
            remote_shell=tosd.get('remote_shell','').replace('$#device#$',dv)

            remote_dir=tosd.get('remote_dir','')

            os.system(remote_init)

            target_exe=os.path.basename(pf)
            pff=remote_dir+'/'+target_exe

            s=remote_push.replace('$#file1#$',pf).replace('$#file2#$',pff)
            os.system(s)

            s=remote_shell+' '+tosd.get('set_executable','')+' '+pff
            os.system(s)

            s=remote_shell+' '+pff
        else:
            s=r['bat'].rstrip('\n')+' '+hosd.get('env_separator','')+' '+pf

            if tosd.get('use_bash_to_run','')!='':
               s=tosd['use_bash_to_run'].replace('$#cmd#$', s)

        s+=' > '+ftmp

        if i.get('sudo','')=='yes':
           s=hosd.get('sudo_pre','')+s

        os.system(s)

        er=i.get('exchange_repo','')
        esr=i.get('exchange_subrepo','')
        el=i.get('exchange_locally','')
        if el!='yes' and er=='': 
           er=ck.cfg['default_exchange_repo_uoa']
           esr=ck.cfg['default_exchange_subrepo_uoa']
        xn=prop.get('name','')

        r=ck.load_text_file({'text_file':ftmp, 'split_to_list':'yes', 'delete_after_read':'yes'})
        if r['return']==0:
           ll=r['lst']

           for l in ll:
               if l=='':
                  # Process if features are not empty
                  if len(prop_id)>0:
                     fuoa=''
                     fuid=''

                     # Exchanging info #################################################################
                     if ex=='yes':
                        if o=='con':
                           ck.out('')
                           ck.out('Exchanging information with '+er+' repository ...')

                        ii={'action':'exchange',
                            'module_uoa':cfg['module_deps']['platform'],
                            'sub_module_uoa':work['self_module_uid'],
                            'repo_uoa':er,
                            'data_name':prop.get('name',''),
                            'extra_info':einf,
                            'all':'yes',
                            'dict':{'features':prop, 'features_misc':prop_all}} 
                        if esr!='': ii['remote_repo_uoa']=esr
                        r=ck.access(ii)
                        if r['return']>0: return r

                        fuoa=r.get('data_uoa','')
                        fuid=r.get('data_uid','')

                        prop=r['dict'].get('features',{})
                        prop_all=r['dict'].get('features_misc',{})

                        if o=='con' and r.get('found','')=='yes':
                           ck.out('  GPGPU CK entry already exists ('+fuid+') - loading latest meta (features) ...')

                     if tp=='cuda':
                        cc=prop_all.get('gpu compute capability','')

                        cc1=''
                        cc2=''
                        cc3=''

                        if cc=='1.0': 
                           cc1='compute_10'
                           cc2='sm_10'
                           cc3='Tesla'
                        elif cc=='1.1': 
                           cc1='compute_11'
                           cc2='sm_11'
                           cc3='Tesla'
                        elif cc=='1.2': 
                           cc1='compute_12'
                           cc2='sm_12'
                           cc3='Tesla'
                        elif cc=='1.3': 
                           cc1='compute_13'
                           cc2='sm_13'
                           cc3='Tesla'
                        elif cc=='2.0': 
                           cc1='compute_20'
                           cc2='sm_20'
                           cc3='Fermi'
                        elif cc=='2.1': 
                           cc1='compute_20'
                           cc2='sm_21'
                           cc3='Fermi'
                        elif cc=='3.0': 
                           cc1='compute_30'
                           cc2='sm_30'
                           cc3='Kepler'
                        elif cc=='3.2': 
                           cc1='compute_32'
                           cc2='sm_32'
                           cc3='Kepler'
                        elif cc=='3.5': 
                           cc1='compute_35'
                           cc2='sm_35'
                           cc3='Kepler'
                        elif cc=='3.7': 
                           cc1='compute_37'
                           cc2='sm_37'
                           cc3='Kepler'
                        elif cc=='5.0': 
                           cc1='compute_50'
                           cc2='sm_50'
                           cc3='Maxwell'
                        elif cc=='5.2': 
                           cc1='compute_52'
                           cc2='sm_52'
                           cc3='Maxwell'
                        elif cc=='5.3': 
                           cc1='compute_53'
                           cc2='sm_53'
                           cc3='Maxwell'
                        elif cc=='6.0': 
                           cc1='compute_60'
                           cc2='sm_60'
                           cc3='Pascal'
                        elif cc=='6.1': 
                           cc1='compute_61'
                           cc2='sm_61'
                           cc3='Pascal'
                        elif cc=='6.2': 
                           cc1='compute_62'
                           cc2='sm_62'
                           cc3='Pascal'
                        elif cc=='7.0': 
                           cc1='compute_70'
                           cc2='sm_70'
                           cc3='Volta'
                        elif cc=='7.1': 
                           cc1='compute_71'
                           cc2='sm_71'
                           cc3='Volta'
                        else:
                           return {'return':1, 'error':'Compute capability of your CUDA device is not recognized ('+cc+') - please report to the authors or update module:platform.gpgpu'}

                        prop_all['compute_capability1']=cc1
                        prop_all['compute_capability2']=cc2
                        prop_all['compute_capability3']=cc3

                        prop['microarchitecture']=cc3

                     jj={"gpgpu":prop, "gpgpu_id":prop_id, "gpgpu_misc":prop_all}

                     if fuoa!='' or fuid!='':
                        jj['gpgpu_uoa']=fuoa
                        jj['gpgpu_misc_uid']=fuid

                     # Print without deps if needed
                     if o=='con':
                        ck.out('')
                        ck.out('Features (properties) in JSON:')

                        r=ck.dumps_json({'dict':jj})
                        if r['return']>0: return r
                        s=r['string']

                        ck.out('')
                        ck.out(s)

                     jj['gpgpu_deps']=deps

                     props.append(jj)

                  # Refresh
                  prop={}
                  prop_id={}
                  prop_all={}

               # Process features
               lx=[]
               if l!='':
                  lx=l.split(':')

               if len(lx)>1:
                  k=lx[0].strip().lower()
                  v=lx[1].strip()

                  if tp=='cuda':
                     if k=='gpu device id':
                        prop_id['gpgpu_device_id']=v
                     elif k=='gpu name':
                        prop['name']=v
                        prop['type']=tp
                     else:
                        prop_all[k]=v
                  else:
                     if k=='platform id':
                        prop_id['gpgpu_platform_id']=v
                     elif k=='device id':
                        prop_id['gpgpu_device_id']=v
                     elif k=='device':
                        prop['name']=v
                        prop['type']=tp
                     elif k=='vendor':
                        prop['vendor']=v
                     else:
                        prop_all[k]=v

    # Check if need to select device and platform
    rr={'return':0, 'features':{'gpgpu':props}}

    if len(props)==0 and o=='con':
       ck.out('')
       ck.out('WARNING: no CK-enabled GPGPU devices found ...')

    if i.get('select','')=='yes' and tp!='':
        # Get Cartesian product of devies and platforms
        platform_ids=[]
        device_ids=[]
        targets=[]
        choices=[]

        # Check if pre-selected
        p_id=i.get('compute_platform_id','')
        d_id=i.get('compute_device_id','')

        for p in props:
            xp_id=p.get('gpgpu_id',{}).get('gpgpu_platform_id','')
            xd_id=p.get('gpgpu_id',{}).get('gpgpu_device_id','')

            if xp_id!='' and xp_id not in platform_ids:
                platform_ids.append(xp_id)

            if xd_id!='' and xd_id not in device_ids:
                device_ids.append(xd_id)

            if (p_id=='' or p_id==xp_id) and (d_id=='' or d_id==xd_id):
                targets.append(p)

                s=p.get('gpgpu',{}).get('name','')

                x=p.get('gpgpu_misc',{}).get('parallel compute units','')
                if x!='':
                   s+=' ; '+x+' compute units '

                x=p.get('gpgpu_misc',{}).get('hardware (device) version','')
                if x!='':
                   s+=' ; '+x

                choices.append(s)

        # Check what to select
        if len(choices)==0:
            return {'return':1, 'error':'can\'t find GPGPU target with specified IDs'}
        elif len(choices)==1 or quiet=='yes' or o!='con':
            ch=0
        else:
            ck.out('')
            ck.out('Multiple GPGPU targets available:')
            ck.out('')

            r=ck.access({'action':'select_list',
                         'module_uoa':cfg['module_deps']['choice'],
                         'choices':choices})
            if r['return']>0: return r
            ch=r['position']

        ich=int(ch)

        p_id=targets[ich].get('gpgpu_id',{}).get('gpgpu_platform_id','')
        d_id=targets[ich].get('gpgpu_id',{}).get('gpgpu_device_id','')

        if o=='con':
            ck.out('')
            ck.out('Selected GPGPU target: '+choices[ich])

        rr['choices']={'compute_number':ich}

        if p_id!='':
            rr['choices']['compute_platform_id']=p_id

        if d_id!='':
            rr['choices']['compute_device_id']=d_id

    os.chdir(pcur)

    return rr

##############################################################################
# viewing entries as html

def show(i):
    """
    Input:  {
              data_uoa
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html         - generated HTML
            }

    """


    h='<h2>GPGPUs (CUDA/OpenCL) participating in crowd-tuning</h2>\n'

    h+='<i>Reuse/extend <a href="https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform.gpgpu">CK JSON meta information</a> of these GPGPU accelerators using "ck pull repo:ck-crowdtuning-platforms" ...</i><br><br>\n'

    h+='<table class="ck_table" border="0" cellpadding="6" cellspacing="0">\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    h+=' <tr style="background-color:#cfcfff;">\n'
    h+='  <td><b>\n'
    h+='   #\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Vendor\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Name\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Type\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':">CK UID</a>\n'
    h+='  </b></td>\n'
    h+=' </tr>\n'

    ruoa=i.get('repo_uoa','')
    muoa=work['self_module_uoa']
    duoa=i.get('data_uoa','')

    r=ck.access({'action':'search',
                 'module_uoa':muoa,
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'add_info':'yes',
                 'add_meta':'yes'})
    if r['return']>0: 
       return {'return':0, 'html':'Error: '+r['error']}

    lst=r['lst']

    num=0
    for q in sorted(lst, key = lambda x: (x.get('meta',{}).get('features',{}).get('vendor','').upper(), \
                                          x.get('meta',{}).get('features',{}).get('name','').upper())):

        num+=1

        duoa=q['data_uoa']
        duid=q['data_uid']

        meta=q['meta']
        ft=meta.get('features',{})
        
        name=ft.get('name','')
        vendor=ft.get('vendor','')
        tp=ft.get('type','')

        h+=' <tr>\n'
        h+='  <td valign="top">\n'
        h+='   '+str(num)+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+vendor+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+name+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+tp+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':'+duid+'">'+duid+'</a>\n'
        h+='  </td>\n'
        h+=' </tr>\n'

    h+='</table><br><br>\n'

    return {'return':0, 'html':h}



#
# Collective Knowledge (compiler choices)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# extract choices to universal pipeline tuning

def extract_to_pipeline(i):
    """
    Input:  {
              (data_uoa) - compiler description
              (file_in)  - JSON pipeline template
              (file_out) - output prepared pipeline to this file
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              pipeline     - prepared pipeline
            }

    """

    fo=i.get('file_out','')
    fi=i.get('file_in','')
    duoa=i.get('data_uoa','')

    # Prepare pipeline template
    ck.out('Preparing pipeline template ...')

    if fi=='':
       fi=os.path.join(work['path'],cfg['pipeline_template'])

    r=ck.load_json_file({'json_file':fi})
    if r['return']>0: return r
    pipeline=r['dict']

    # Load or select compiler description
    ck.out('')
    ck.out('Selecting compiler and description ...')
    if duoa=='':
       rx=ck.access({'action':'search',
                     'module_uoa':work['self_module_uid']})
       if rx['return']>0: return rx
       lst=rx['lst']

       if len(lst)==0:
          duoa=''
       elif len(lst)==1:
          duid=lst[0]['data_uid']
          duoa=lst[0]['data_uoa']
       else:
          # SELECTOR *************************************
          ck.out('')
          ck.out('Multiple choices available:')
          ck.out('')
          r=ck.select_uoa({'choices':lst})
          if r['return']>0: return r
          duoa=r['choice']

    if duoa=='':
       return {'return':1, 'error':'no compiler description selected'}

    # Load compiler desc
    rx=ck.access({'action':'load',
                  'module_uoa':work['self_module_uid'],
                  'data_uoa':duoa})
    if rx['return']>0: return rx
    d=rx['dict']
    dsc=rx.get('desc',{})

    dx=dsc.get('all_compiler_flags_desc',{})

    # Update pipeline
    ck.out('')
    ck.out('Updating pipeline choices with compiler flags ...')

    if 'pipeline' not in pipeline: pipeline['pipeline']={}
    px=pipeline['pipeline']

    px['choices_desc']={}
    for q in dx:
        qq=dx[q]
        q1=q
        if q1.startswith('##'): q1=q1[1:]
        q1='##compiler_flags'+q1
        px['choices_desc'][q1]=qq

    # Saving file
    if fo!='':
       ck.out('')
       ck.out('Writing pipeline ...')

       rx=ck.save_json_to_file({'json_file':fo, 'dict':pipeline})
       if rx['return']>0: return rx

    return {'return':0, 'pipeline':pipeline}

##############################################################################
# extract optimization flags and parameters from a compiler for autotuning (currently GCC)
# Note: GCC sources are needed - this script searches for GCC opts in all sub-directories
#       in a current directory. Therefore, just untar all GCC sources in current directory.
#
# (partially based on scripts from Abdul Wahid Memon and Yuriy Kashnikov)

def extract_opts(i):
    """
    Input:  {
              (record)               - if 'yes, record to repo
              (repo_uoa)             - repo UOA to record compiler description
              (data_uoa)             - data UOA to record compiler description (if empty, autogenerate)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import sys
    import datetime

    o=i.get('out','')

    p=os.getcwd()

    f1a=cfg['gcc_src_opt_file1']
    f1b=cfg['gcc_src_opt_file2']
    f2=cfg['gcc_src_params']

    results={}

    # Checking if GCC 
    try:
       dirList=os.listdir(p)
    except Exception as e:
        None
    else:
        for fn in dirList:
            pp=os.path.join(p, fn)
            if os.path.isdir(pp) and fn.startswith('gcc-'):
               found=False

               p1a=os.path.join(pp,f1a)
               p1b=os.path.join(pp,f1b)
               p2=os.path.join(pp,f2)

               if os.path.isfile(p1a) or os.path.isfile(p1b) or os.path.isfile(p2):
                  # Found GCC source directory with needed files
                  ck.out('*************************************')
                  ck.out('Processing GCC directory: '+fn)

                  ck.out('')

                  ver=fn[4:]
                  lver=ver.split('.')

                  dd={}

                  iopt=0
                  iparam=0

                  p1=''
                  if os.path.isfile(p1a):
                     p1=p1a
                  elif os.path.isfile(p1b):
                     p1=p1b

                  if p1!='':
                     # Load opt file
                     rx=ck.load_text_file({'text_file':p1,
                                           'split_to_list':'yes', 
                                           'encoding':sys.stdin.encoding})
                     if rx['return']>0: return rx
                     lopts=rx['lst']

                     found=False
                     for q in range(0, len(lopts)):
                         qq=lopts[q].strip()

                         if not found:
                            if qq=='@item Optimization Options':
                               ck.out('Found optimization flags')
                               found=True
                         else:
                            if qq.startswith('@end ') or qq=='':
                               break

                            if qq.startswith('@gccoptlist{'):
                               qq=qq[11:]
                            elif qq.startswith('@'):
                               continue                               

                            if qq.endswith('@gol'):
                               qq=qq[:-5]

                            jj=qq.split(' ')

                            # Check if base flag
                            if len(jj)>0 and jj[0].startswith('-O'):
                               choice=[]
                               if '-O3' in jj: choice.append('-O3')
                               for j in jj:
                                   if j!='' and j!='-O3' and j!='-O':
                                      if j.endswith('}'): j=j[:-1].strip()
                                      choice.append(j)

                               dd["##base_opt"]={
                                       "choice": choice,
                                       "default": "", 
                                       "desc": "base compiler flag", 
                                       "sort": 10000, 
                                       "tags": [
                                         "base", 
                                         "basic", 
                                         "optimization"
                                       ], 
                                       "type": "text"
                                     }

                            else:
                               for j in jj:
                                   if j!='' and not j.startswith('--param') and not j.startswith('@') and j.startswith('-f') and \
                                                j.find('profile')==-1 and j.find('coverage')==-1:
                                      if '@' in j:
                                         iparam+=1

                                         ij=j.find('@')
                                         opt=j[:ij]

                                         if opt.endswith('@var{n}'): opt=opt[:-7]

                                         ck.out('Adding param '+str(iparam)+' '+opt)

                                         dd['##param-'+opt]={
                                           "can_omit": "yes", 
                                           "default": "", 
                                           "desc": "compiler flag: "+j, 
                                           "sort": iparam*10+30000, 
                                           "explore_prefix": j, 
                                           "explore_start": 0, 
                                           "explore_step": 1, 
                                           "explore_stop": 0, 
                                           "tags": [
                                             "basic", 
                                             "parametric",
                                             "optimization"
                                           ], 
                                           "type":"integer"
                                         }

                                      else:
                                         iopt+=1

                                         opt=j[2:]

                                         if opt.startswith('no-'):
                                            opt=opt[3:]

                                         ck.out('Adding opt '+str(iopt)+' '+j)

                                         dd['##bool-'+opt]={
                                           "can_omit": "yes", 
                                           "choice": [
                                             "-f"+opt, 
                                             "-fno-"+opt
                                           ], 
                                           "default": "", 
                                           "desc": "compiler flag: "+j, 
                                           "sort": iopt*10+10000, 
                                           "tags": [
                                             "basic", 
                                             "boolean",
                                             "optimization"
                                           ], 
                                           "type":"text"
                                         }

                  # Checking params
                  if os.path.isfile(p2):
                     # Load opt file
                     rx=ck.load_text_file({'text_file':p2,
                                           'split_to_list':'yes', 
                                           'encoding':sys.stdin.encoding})
                     if rx['return']>0: return rx
                     lparams=rx['lst']

                     # Parsing params
                     opt=''
                     opt1=''
                     desc=''
                     desc1=''

                     add=False
                     finish=False

                     for q in range(0, len(lparams)):
                         qq=lparams[q].strip()
                         if qq.startswith('DEFPARAM'):
                            iparam+=1

                            q+=1
                            opt=lparams[q].strip()[1:-2]

                            if opt.find('profile')==-1 and opt.find('coverage')==-1:
                               q+=1
                               desc=lparams[q].strip()[1:-2]
                               line='x'
                               while True:
                                  q+=1
                                  line=lparams[q].strip()
                                  if line[-1]==')': break
                                  desc+=line[1:-2]

                               e1=0
                               e2=0

                               exp=line[:-1].split(',')
                               for x in exp[:]:
                                   if (x.rfind('\"') != -1):
                                      exp.remove(x)

                               skip=False
                               for j in range(0, len(exp)):
                                   jj=exp[j].strip()
                                   if jj.find('*')>0 or jj.find('_')>0:
                                      skip=True
                                   else:
                                      jj=int(exp[j].strip())
                                      exp[j]=jj

                               if not skip:
                                  if len(exp)>1 and exp[2]>exp[1]:
                                     e1=exp[1]
                                     e2=exp[2]
                                  else:
                                     e1=0
                                     e2=exp[0]*2

                                  ck.out('Adding param '+str(iparam)+' "'+opt+'" - '+desc)

                                  dd['##param-'+opt]={
                                    "can_omit": "yes", 
                                    "default": "", 
                                    "desc": "compiler flag: --param "+opt+"= ("+desc+")", 
                                    "sort": iparam*10+30000, 
                                    "explore_prefix": "--param "+opt+"=", 
                                    "explore_start": e1, 
                                    "explore_step": 1, 
                                    "explore_stop": e2, 
                                    "tags": [
                                      "basic", 
                                      "parametric",
                                      "optimization"
                                    ], 
                                    "type":"integer"
                                  }

                  # Prepare CK entry
                  if i.get('record','')=='yes':
                     duoa=i.get('data_uoa','')
                     if duoa=='':
                        duoa=fn+'-auto'

                     if o=='con':
                        ck.out('')
                        ck.out('Recording to '+duoa+' ...')

                     ii={'action':'add',
                         'module_uoa':work['self_module_uid'],
                         'repo_uoa':i.get('repo_uoa',''),
                         'data_uoa':duoa,
                         'desc':{'all_compiler_flags_desc':dd},
                         'dict':{
                           "tags": [
                             "compiler", 
                             "gcc", 
                             "v"+lver[0], 
                             "v"+lver[0]+"."+lver[1],
                             "auto"
                           ]
                         }
                        }
                     r=ck.access(ii)
                     if r['return']>0: return r

                  p9=os.path.join(pp, 'ChangeLog')
                  year=''
                  if os.path.isfile(p9):
                     t = os.path.getmtime(p9)
                     t1=str(datetime.datetime.fromtimestamp(t))
                     year=t1[:4]

                  results[fn]={'boolean_opts':iopt, 'parametric_opts':iparam, 'year':year}

                  # Final info
                  if o=='con':
                     ck.out('')
                     ck.out('Compiler:                     '+str(fn))
                     ck.out('  Year:                       '+year)
                     ck.out('  Number of boolean opts:     '+str(iopt))
                     ck.out('  Number of parameteric opts: '+str(iparam))

    # Summary
    if len(results)>0:
       ck.out('*********************************************************')
       for q in sorted(list(results.keys())):
           qq=results[q]

           x=qq['year']+', '

           ib=str(qq['boolean_opts'])
           x+=' '*(6-len(ib))+ib+', '

           ip=str(qq['parametric_opts'])
           x+=' '*(6-len(ip))+ip+', '

           x+=q+' '*(10-len(q))

           ck.out(x)

    return {'return':0}

##############################################################################
# extract optimization flags and parameters from a compiler for autotuning (currently GCC)

def extract_opts_new(i):
    """
    Input:  {
              (host_os)              - host OS (detect, if omitted)
              (target_os)            - OS module to check (if omitted, analyze host)
              (device_id)            - device id if remote (such as adb)

              (compiler_id)          - currently support only "gcc" (default)

              (repo_uoa)             - repo UOA to record compiler description
              (data_uoa)             - data UOA to record compiler description (if empty, autogenerate)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import sys

    o=i.get('out','')

    comp_id=i.get('compiler_id','').lower()
    if comp_id=='': comp_id='gcc'

    # Check OS
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')

    # Get some info about platforms
    ii={'action':'detect',
        'module_uoa':cfg['module_deps']['platform.os'],
        'host_os':hos,
        'target_os':tos,
        'device_id':tdid}
    r=ck.access(ii)
    if r['return']>0: return r

    hos=r['host_os_uid']
    hosx=r['host_os_uoa']
    hosd=r['host_os_dict']

    tos=r['os_uid']
    tosx=r['os_uoa']
    tosd=r['os_dict']

    tdid=r['device_id']

    sext=hosd.get('script_ext','')
    scall=hosd.get('env_call','')
    sbp=hosd.get('bin_prefix','')

    tags='compiler'
    if comp_id=='gcc': tags+=',gcc'

    # Get compiler env
    ii={'action':'set',
        'module_uoa':cfg['module_deps']['env'],
        'host_os':hos,
        'target_os':tos,
        'device_id':tdid,
        'tags':tags,
        'out':o}
    r=ck.access(ii)
    if r['return']>0: return r

    env_uoa=r['env_uoa']
    bat=r['bat']

    # Detect version
    ii={'action':'internal_detect',
        'module_uoa':cfg['module_deps']['soft'],
        'host_os':hos,
        'target_os':tos,
        'device_id':tdid,
        'tags':tags,
        'env':bat}
    r=ck.access(ii)
    if r['return']>0: return r

    vstr=r['version_str']
    if o=='con':
       ck.out('Detected GCC version: '+vstr)

    # Prepare batch file
    rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':sext, 'remove_dir':'yes'})
    if rx['return']>0: return rx
    fbat=rx['file_name']

    # Prepare 2 tmp files
    rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.tmp', 'remove_dir':'yes'})
    if rx['return']>0: return rx
    fout1=rx['file_name']

    # Prepare GCC help
    bat+='\n'
    bat+='gcc --help=optimizer > '+fout1+'\n'

    # Save file
    rx=ck.save_text_file({'text_file':fbat, 'string':bat})
    if rx['return']>0: return rx

    # Execute 
    y=scall+' '+sbp+fbat

    if o=='con':
       ck.out('')
       ck.out('Executing prepared batch file '+fbat+' ...')
       ck.out('')

    sys.stdout.flush()
    rx=os.system(y)

    os.remove(fbat)

    # Load opt file
    rx=ck.load_text_file({'text_file':fout1,
                          'split_to_list':'yes', 
                          'encoding':sys.stdin.encoding,
                          'delete_after_read':'yes'})
    if rx['return']>0: return rx
    lopts=rx['lst']

    # Check if want params
    ck.out('')
    r=ck.inp({'text':'Enter full path to params.def file if you have GCC sources (or Enter to skip it): '})
    if r['return']>0: return r

    fpar=r['string'].strip()

    lparams=[]
    if fpar!='':
       rx=ck.load_text_file({'text_file':fpar,
                             'split_to_list':'yes', 
                             'encoding':sys.stdin.encoding})
       if rx['return']>0: return rx
       lparams=rx['lst']

    # Parsing opts
    dd={"##base_opt": {
            "choice": [
              "-O3", 
              "-Ofast", 
              "-O0", 
              "-O1", 
              "-O2", 
              "-Os"
            ], 
            "default": "", 
            "desc": "base compiler flag", 
            "sort": 10000, 
            "tags": [
              "base", 
              "basic", 
              "optimization"
            ], 
            "type": "text"
          }
       }

    iopt=0
    iparam=0

    opt=''
    opt1=''
    desc=''
    desc1=''

    add=False
    finish=False

    for q in range(1, len(lopts)):
        qq=lopts[q]
        if len(qq)>2:
           if qq[2]=='-':
              qq=qq.strip()
              j=qq.find(' ')
              desc1=''
              if j>=0:
                 desc1=qq[j:].strip()
              else:
                 j=len(qq)
              opt1=qq[1:j]

              if not opt1.startswith('O'):
                 if opt=='': 
                    opt=opt1
                    desc=desc1
                 else:
                    add=True

           else:
              qq=qq.strip()
              if len(qq)>0:
                 desc+=' '+qq
        else:
           add=True
           finish=True

        if add and opt!='':
           iopt+=1

           opt=opt[1:]
           if opt.startswith('fno-'):
              opt=opt[4:]

           ck.out('Adding opt '+str(iopt)+' "'+opt+'" - '+desc)

           dd['##'+opt]={
             "can_omit": "yes", 
             "choice": [
               "-f"+opt, 
               "-fno-"+opt
             ], 
             "default": "", 
             "desc": "compiler flag: -f"+opt+"("+desc+")", 
             "sort": iopt*100, 
             "tags": [
               "basic", 
              "optimization"
             ], 
             "type":"text"
           }

           opt=opt1 	
           desc=desc1

           add=False

        if finish: 
           break

    # Parsing params
    opt=''
    opt1=''
    desc=''
    desc1=''

    add=False
    finish=False

    for q in range(0, len(lparams)):
        qq=lparams[q].strip()
        if qq.startswith('DEFPARAM'):
           iparam+=1

           q+=1
           opt=lparams[q].strip()[1:-2]

           q+=1
           desc=lparams[q].strip()[1:-2]
           line='x'
           while True:
              q+=1
              line=lparams[q].strip()
              if line[-1]==')': break
              desc+=line[1:-2]

           e1=0
           e2=0

           exp=line[:-1].split(',')
           for x in exp[:]:
               if (x.rfind('\"') != -1):
                  exp.remove(x)

           skip=False
           for j in range(0, len(exp)):
               jj=exp[j].strip()
               if jj.find('*')>0 or jj.find('_')>0:
                  skip=True
               else:
                  jj=int(exp[j].strip())
                  exp[j]=jj

           if not skip:
              if exp[2]>exp[1]:
                 e1=exp[1]
                 e2=exp[2]
              else:
                 e1=0
                 e2=exp[0]*2

              ck.out('Adding param '+str(iparam)+' "'+opt+'" - '+desc)

              dd['##param_'+opt]={
                "can_omit": "yes", 
                "default": "", 
                "desc": "compiler flag: --param "+opt+"= ("+desc+")", 
                "sort": iparam*100+30000, 
                "explore_prefix": "--param "+opt+"=", 
                "explore_start": e1, 
                "explore_step": 1, 
                "explore_stop": e2, 
                "tags": [
                  "basic", 
                 "optimization"
                ], 
                "type":"integer"
              }

    # Prepare CK entry
    v=vstr.split('.')
    duoa=i.get('data_uoa','')
    if duoa=='':
       duoa='gcc-'+v[0]+v[1]+'x'

    if o=='con':
       ck.out('')
       ck.out('Recording to '+duoa+' ...')

    ii={'action':'add',
        'module_uoa':work['self_module_uid'],
        'repo_uoa':i.get('repo_uoa',''),
        'data_uoa':duoa,
        'desc':{'all_compiler_flags_desc':dd},
        'dict':{
          "tags": [
          "compiler", 
          "gcc", 
          "v"+v[0], 
          "v"+v[0]+"."+v[1]
          ]
        }
       }
    r=ck.access(ii)
    if r['return']>0: return r

    # Final info
    if o=='con':
       ck.out('')
       ck.out('Compiler version:           '+vstr)
       ck.out('Number of boolean opts:     '+str(iopt))
       ck.out('Number of parameteric opts: '+str(iparam))

    return {'return':0}

##############################################################################
# viewing compiler description as HTML

def html_viewer(i):
    """
    Input:  {
              data_uoa

              url_base
              url_pull

              url_pull_tmp
              tmp_data_uoa

              url_wiki

              html_share

              form_name     - current form name

              (all_params)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html         - generated HTML

              raw          - if 'yes', output in CK raw format
              show_top     - if 'yes', include CK top header with QR-code
            }

    """

    h=''
    raw='no'
    top='yes'

    duoa=i['data_uoa']

    if duoa!='':
       # Load entry
       rx=ck.access({'action':'load',
                     'module_uoa':work['self_module_uid'],
                     'data_uoa':duoa})
       if rx['return']>0: return rx

       desc=rx['desc']

       h+='<hr>\n'

       # Output compiler flags description
       acfd=desc.get('all_compiler_flags_desc',{})
       if len(acfd)>0:
          h+='<center>\n'
          h+='<b>Compiler optimization flag description (boolean and parametric)</b><br><br>\n'

          h+='<table border="1" cellpadding="5" cellspacing="0">\n'

          h+=' <tr>\n'
          h+='  <td><b>JSON key</b></td>\n'
          h+='  <td><b>Description</b></td>\n'
          h+='  <td><b>Type</b></td>\n'
          h+='  <td><b>Order</b></td>\n'
          h+='  <td><b>Choices</b></td>\n'
          h+='  <td><b>Parameter prefix</b></td>\n'
          h+='  <td><b>Start</b></td>\n'
          h+='  <td><b>Stop</b></td>\n'
          h+='  <td><b>Step</b></td>\n'
          h+='  <td><b>Tags</b></td>\n'
          h+=' </tr>\n'

          ll=list(acfd.keys())
          for k in sorted(ll, key=lambda j: (int(acfd[j].get('sort',0)))):
              v=acfd[k]

              tags=v.get('tags',[])
              stags=''
              for q in tags:
                  if stags!='': stags+=', '
                  stags+=q

              desc=v.get('desc','')
              sort=v.get('sort','')

              tp=v.get('type','')

              choice=v.get('choice',[])
              schoice=''
              for q in choice:
                  if schoice!='': schoice+=', '
                  schoice+=q

              ep=v.get('explore_prefix','')
              estart=str(v.get('explore_start',''))
              estep=str(v.get('explore_step',''))
              estop=str(v.get('explore_stop',''))

              h+=' <tr>\n'
              h+='  <td><i>'+k+'</i></td>\n'
              h+='  <td>'+desc+'</td>\n'
              h+='  <td>'+tp+'</td>\n'
              h+='  <td>'+str(sort)+'</td>\n'
              h+='  <td>'+schoice+'</td>\n'
              h+='  <td>'+ep+'</td>\n'
              h+='  <td>'+estart+'</td>\n'
              h+='  <td>'+estop+'</td>\n'
              h+='  <td>'+estep+'</td>\n'
              h+='  <td>'+stags+'</td>\n'
              h+=' </tr>\n'

          h+='</table>\n'
          h+='</center>\n'

       h+='<hr>\n'

    return {'return':0, 'raw':raw, 'show_top':top, 'html':h}



#
# Collective Knowledge: autotune any custom dimensions provided by user
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
import copy

line='****************************************************************'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# crowdsource these experiments

def crowdsource(i):
    global cfg, work

    import copy

    mcfg=i.get('module_cfg',{})
    if len(mcfg)>0: 
       cfg=mcfg

    mwork=i.get('module_work',{})
    if len(mwork)>0: work=mwork

    # Setting output
    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    quiet=i.get('quiet','')

    er=i.get('exchange_repo','')
    if er=='': er=ck.cfg['default_exchange_repo_uoa']
    esr=i.get('exchange_subrepo','')
    if esr=='': esr=ck.cfg['default_exchange_subrepo_uoa']

    if i.get('local','')=='yes': 
       er='local'
       esr=''
    else:
       return {'return':1, 'error':'can\'t crowdsource these experiments'}

    la=i.get('local_autotuning','')

    # Get user 
    user=''

    mcfg={}
    ii={'action':'load',
        'module_uoa':'module',
        'data_uoa':cfg['module_deps']['program.optimization']}
    r=ck.access(ii)
    if r['return']==0:
       mcfg=r['dict']

       dcfg={}
       ii={'action':'load',
           'module_uoa':mcfg['module_deps']['cfg'],
           'data_uoa':mcfg['cfg_uoa']}
       r=ck.access(ii)
       if r['return']>0 and r['return']!=16: return r
       if r['return']!=16:
          dcfg=r['dict']

       user=dcfg.get('user_email','')

    ceuoa=i.get('compiler_env_uoa', '')

    if ceuoa!='':
       rx=ck.access({'action':'load',
                     'module_uoa':cfg['module_deps']['env'],
                     'data_uoa':ceuoa})
       if rx['return']>0: return rx
       ceuoa=rx['data_uid']

    # Initialize local environment for program optimization ***********************************************************
    pi=i.get('platform_info',{})
    if len(pi)==0:
       ii=copy.deepcopy(i)
       ii['action']='initialize'
       ii['module_uoa']=cfg['module_deps']['program.optimization']
       ii['exchange_repo']=er
       ii['exchange_subrepo']=esr
       r=ck.access(ii)
       if r['return']>0: return r

       pi=r['platform_info']
       user=r.get('user','')

    hos=pi['host_os_uoa']
    hosd=pi['host_os_dict']

    tos=pi['os_uoa']
    tosd=pi['os_dict']
    tbits=tosd.get('bits','')

    remote=tosd.get('remote','')

    tdid=pi['device_id']

    program_tags=i.get('program_tags','')
    if program_tags=='' and i.get('local_autotuning','')!='yes' and i.get('data_uoa','')=='':
       program_tags=cfg['program_tags']

    # Check that has minimal dependencies for this scenario ***********************************************************
    sdeps=i.get('dependencies',{}) # useful to preset inside crowd-tuning
    if len(sdeps)==0:
       sdeps=copy.deepcopy(cfg.get('deps',{}))
    if len(sdeps)>0:
       if o=='con':
          ck.out(line)
          ck.out('Resolving software dependencies required for this scenario ...')
          ck.out('')

       if ceuoa!='':
          x=sdeps.get('compiler',{})
          if len(x)>0:
             if 'cus' in x: del(x['cus'])
             if 'deps' in x: del(x['deps'])
             x['uoa']=ceuoa
             sdeps['compiler']=x

       ii={'action':'resolve',
           'module_uoa':cfg['module_deps']['env'],
           'host_os':hos,
           'target_os':tos,
           'device_id':tdid,
           'deps':sdeps,
           'add_customize':'yes'}
       if quiet=='yes': 
          ii['random']='yes'
       else:
          ii['out']=oo
       rx=ck.access(ii)
       if rx['return']>0: return rx

       sdeps=rx['deps'] # Update deps (add UOA)

    cpu_name=pi.get('features',{}).get('cpu',{}).get('name','')
    compiler_soft_uoa=sdeps.get('compiler',{}).get('dict',{}).get('soft_uoa','')
    compiler_env=sdeps.get('compiler',{}).get('bat','')

    plat_extra={}
    pft=pi.get('features',{})
    for q in pft:
        if q.endswith('_uid'):
           plat_extra[q]=pft[q]
        elif type(pft[q])==dict and pft[q].get('name','')!='':
           plat_extra[q+'_name']=pft[q]['name']

    # Detect real compiler version ***********************************************************
    compiler_version=''
    compiler=''
    if compiler_soft_uoa!="":
       if o=='con':
          ck.out(line)
          ck.out('Detecting compiler version ...')

       ii={'action':'internal_detect',
           'module_uoa':cfg['module_deps']['soft'],
           'data_uoa':compiler_soft_uoa,
           'host_os':hos,
           'target_os':tos,
           'target_device_id':tdid,
           'env':compiler_env}
       r=ck.access(ii)
       if r['return']>0: return r

       compiler_version=r['version_str']

       compiler=cfg.get('compiler_name','')+' '+compiler_version

    if o=='con':
       ck.out('')
       if compiler!='':
          ck.out('* Compiler: '+compiler)
       ck.out('* CPU:      '+cpu_name)
    
    # Start preparing input to run program.optimization
    ii=copy.deepcopy(i)

    em={'cpu_name':cpu_name}
    if compiler!='': em['compiler']=compiler

    ii['action']='run'
    ii['module_uoa']=cfg['module_deps']['program.optimization']

    ii['host_os']=hos
    ii['target_os']=tos
    ii['target_device_id']=tdid
    ii['dependencies']=sdeps

    ii['scenario_cfg']=cfg

    ii['platform_info']=pi

    ii['program_tags']=program_tags

    ii['scenario_module_uoa']=work['self_module_uid']

    ii['experiment_meta']=em

    ii['experiment_meta_extra']=plat_extra

    ii['exchange_repo']=er
    ii['exchange_subrepo']=esr

    ii['user']=user

    # Select sub-scenario ********************************************************************
    from random import randint
    ss=1 # num of scenarios

    sx=randint(1,ss)

    rr={'return':0}

    if sx==1 or la=='yes':
       # **************************************************************** explore random program/dataset
       sdesc='custom dimensions autotuner'
       if o=='con':
          ck.out('')
          ck.out('  ****** Sub-scenario: '+sdesc+' ******')

       ii['subscenario_desc']=sdesc

       rr=ck.access(ii)
       if rr['return']>0: return rr

    rr['platform_info']=pi

    return rr

##############################################################################
# view solutions in html

def html_viewer(i):
    """
    See in module "experiment.tune.compiler.flags"
    """

    i['module_uoa']='experiment.tune.compiler.flags'
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)

##############################################################################
# replay optimization

def replay(i):
    """
    See in module "program.optimization"
    """

    i['module_uoa']=cfg['module_deps']['program.optimization']
    i['module_ref_uoa']=work['self_module_uid']
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)

##############################################################################
# prune compiler flags to find minimal set of choices

def prune(i):
    """
    See in module "program.optimization"
    """

    return {'return':1, 'error':'pruning is not yet supported in this scenario'}



#
# Collective Knowledge (indexing through ElasticSearch)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

cfg = {}  # Will be updated by CK (meta description of this module)
work = {}  # Will be updated by CK (temporal data)
ck = None  # Will be updated by CK (initialized CK kernel)

# Local settings

##############################################################################
# Initialize module


def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return': 0}

##############################################################################
# turn indexing on


def on(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    i['status'] = 'yes'

    r = status(i)
    if r['return'] > 0:
        return r

    if o == 'con':
        ck.out('Indexing is on')

    return {'return': 0}

##############################################################################
# turn indexing off


def off(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    i['status'] = 'no'

    r = status(i)
    if r['return'] > 0:
        return r

    if o == 'con':
        ck.out('Indexing is off')

    return {'return': 0}

##############################################################################
# show indexing status


def show(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    i['status'] = ''

    r = status(i)
    if r['return'] > 0:
        return r

    s = r['status']
    if s == 'yes':
        sx = 'on'
    else:
        sx = 'off'

    if o == 'con':
        ck.out('Indexing status: '+sx)

    return {'return': 0}

##############################################################################
# check indexing status


def status(i):
    """

    Input:  {
               status - if 'yes', turn it on 
                        if 'no', turn it off 
                        if '', return status
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
              status
            }

    """

    # Get current configuration
    cfg = {}

    r = ck.access({'action': 'load',
                   'repo_uoa': ck.cfg['repo_name_default'],
                   'module_uoa': ck.cfg['subdir_kernel'],
                   'data_uoa': ck.cfg['subdir_kernel_default']})
    if r['return'] == 0:
        cfg.update(r['dict'])

    r = ck.access({'action': 'load',
                   'repo_uoa': ck.cfg['repo_name_local'],
                   'module_uoa': ck.cfg['subdir_kernel'],
                   'data_uoa': ck.cfg['subdir_kernel_default']})
    if r['return'] > 0:
        if r['return'] != 16:
            return r

        ck.out('')
        ck.out('We strongly suggest you to setup local repository first!')
        return {'return': 0}

    cfg.update(r['dict'])

    # Turn on indexing
    st = i.get('status', '')
    s = cfg.get('use_indexing', ck.cfg.get('use_indexing', ''))

    if st != '':
        cfg['use_indexing'] = st
        s = st

    r = ck.access({'action': 'update',
                   'repo_uoa': ck.cfg['repo_name_local'],
                   'module_uoa': ck.cfg['subdir_kernel'],
                   'data_uoa': ck.cfg['subdir_kernel_default'],
                   'dict': cfg,
                   'substitute': 'yes',
                   'ignore_update': 'yes',
                   'skip_indexing': 'yes'})
    if r['return'] > 0:
        return r

    return {'return': 0, 'status': s}

##############################################################################
# check indexing status


def test(i):
    """

    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    r = ck.access_index_server({'request': 'TEST', 'dict': {}})
    if r['return'] > 0:
        return r
    dd = r['dict']

    status = dd.get('status', 0)
    if status != 200:
        return {'return': 1, 'error': 'returned status is not 200'}

    version = dd.get('version', {}).get('number', '')

    if o == 'con':
        ck.out('Indexing server is working (version = '+version+')')

    return r

##############################################################################
# clean whole index


def clean(i):
    """

    Input:  {
              (force)  - if 'yes', force cleaning
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o = i.get('out', '')

    to_delete = True
    if o == 'con' and i.get('force', '') != 'yes':
        r = ck.inp({'text': 'Are you sure to clean the whole index (y/N): '})
        c = r['string'].lower()
        if c != 'y' and c != 'yes':
            to_delete = False

    if to_delete:
        r = ck.access_index_server(
            {'request': 'DELETE', 'path': '/_all', 'dict': {}})
        if r['return'] > 0:
            return r
        dd = r['dict']

        status = dd.get('status', 0)
        err = dd.get('error', '')

        if err != '':
            r = {'return': 1, 'error': err}

    return r
